@@ -1,42 +1,58 @@
+from __future__ import annotations
+
+import asyncio
 import ctypes
 import errno
 import functools
 import logging
 import socket
+import ssl
 import struct
 import sys
 import weakref
 from ssl import SSLCertVerificationError, SSLError
-
-from tornado import gen
-
-try:
-    import ssl
-except ImportError:
-    ssl = None  # type: ignore
+from typing import Any, ClassVar
 
 from tlz import sliding_window
-from tornado import netutil
-from tornado.iostream import StreamClosedError
+from tornado import gen, netutil
+from tornado.iostream import IOStream, StreamClosedError
 from tornado.tcpclient import TCPClient
 from tornado.tcpserver import TCPServer
 
 import dask
 from dask.utils import parse_timedelta
 
-from ..protocol.utils import pack_frames_prelude, unpack_frames
-from ..system import MEMORY_LIMIT
-from ..threadpoolexecutor import ThreadPoolExecutor
-from ..utils import ensure_ip, get_ip, get_ipv6, nbytes
-from .addressing import parse_host_port, unparse_host_port
-from .core import Comm, CommClosedError, Connector, FatalCommClosedError, Listener
-from .registry import Backend
-from .utils import ensure_concrete_host, from_frames, get_tcp_server_address, to_frames
+from distributed.comm.addressing import parse_host_port, unparse_host_port
+from distributed.comm.core import (
+    Comm,
+    CommClosedError,
+    Connector,
+    FatalCommClosedError,
+    Listener,
+)
+from distributed.comm.registry import Backend
+from distributed.comm.utils import (
+    ensure_concrete_host,
+    from_frames,
+    get_tcp_server_address,
+    host_array,
+    to_frames,
+)
+from distributed.protocol.utils import pack_frames_prelude, unpack_frames
+from distributed.system import MEMORY_LIMIT
+from distributed.utils import ensure_ip, ensure_memoryview, get_ip, get_ipv6, nbytes
 
 logger = logging.getLogger(__name__)
 
 
-C_INT_MAX = 256 ** ctypes.sizeof(ctypes.c_int) // 2 - 1
+# Workaround for OpenSSL 1.0.2.
+# Can drop with OpenSSL 1.1.1 used by Python 3.10+.
+# ref: https://bugs.python.org/issue42853
+if sys.version_info < (3, 10):
+    OPENSSL_MAX_CHUNKSIZE = 256 ** ctypes.sizeof(ctypes.c_int) // 2 - 1
+else:
+    OPENSSL_MAX_CHUNKSIZE = 256 ** ctypes.sizeof(ctypes.c_size_t) - 1
+
 MAX_BUFFER_SIZE = MEMORY_LIMIT / 2
 
 
@@ -95,8 +111,8 @@ def set_tcp_timeout(comm):
             logger.debug("Setting TCP user timeout: %d ms", timeout * 1000)
             TCP_USER_TIMEOUT = 18  # since Linux 2.6.37
             sock.setsockopt(socket.SOL_TCP, TCP_USER_TIMEOUT, timeout * 1000)
-    except OSError as e:
-        logger.warning("Could not set timeout on TCP stream: %s", e)
+    except OSError:
+        logger.exception("Could not set timeout on TCP stream.")
 
 
 def get_stream_address(comm):
@@ -120,8 +136,8 @@ def convert_stream_closed_error(obj, exc):
     if exc.real_error is not None:
         # The stream was closed because of an underlying OS error
         exc = exc.real_error
-        if ssl and isinstance(exc, ssl.SSLError):
-            if "UNKNOWN_CA" in exc.reason:
+        if isinstance(exc, ssl.SSLError):
+            if exc.reason and "UNKNOWN_CA" in exc.reason:
                 raise FatalCommClosedError(f"in {obj}: {exc.__class__.__name__}: {exc}")
         raise CommClosedError(f"in {obj}: {exc.__class__.__name__}: {exc}") from exc
     else:
@@ -145,11 +161,14 @@ class TCP(Comm):
     An established communication based on an underlying Tornado IOStream.
     """
 
-    max_shard_size = dask.utils.parse_bytes(dask.config.get("distributed.comm.shard"))
+    max_shard_size: ClassVar[int] = dask.utils.parse_bytes(
+        dask.config.get("distributed.comm.shard")
+    )
+    stream: IOStream | None
 
     def __init__(
         self,
-        stream,
+        stream: IOStream,
         local_addr: str,
         peer_addr: str,
         deserialize: bool = True,
@@ -175,7 +194,9 @@ class TCP(Comm):
         pass
 
     def _get_finalizer(self):
-        def finalize(stream=self.stream, r=repr(self)):
+        r = repr(self)
+
+        def finalize(stream=self.stream, r=r):
             # stream is None if a StreamClosedError is raised during interpreter
             # shutdown
             if stream is not None and not stream.closed():
@@ -204,13 +225,13 @@ class TCP(Comm):
             frames_nbytes = await stream.read_bytes(fmt_size)
             (frames_nbytes,) = struct.unpack(fmt, frames_nbytes)
 
-            frames = memoryview(bytearray(frames_nbytes))
-            # Workaround for OpenSSL 1.0.2 (can drop with OpenSSL 1.1.1)
+            frames = host_array(frames_nbytes)
             for i, j in sliding_window(
-                2, range(0, frames_nbytes + C_INT_MAX, C_INT_MAX)
+                2,
+                range(0, frames_nbytes + OPENSSL_MAX_CHUNKSIZE, OPENSSL_MAX_CHUNKSIZE),
             ):
                 chunk = frames[i:j]
-                chunk_nbytes = len(chunk)
+                chunk_nbytes = chunk.nbytes
                 n = await stream.read_into(chunk)
                 assert n == chunk_nbytes, (n, chunk_nbytes)
         except StreamClosedError as e:
@@ -218,11 +239,12 @@ class TCP(Comm):
             self._closed = True
             if not sys.is_finalizing():
                 convert_stream_closed_error(self, e)
-        except Exception:
-            # Some OSError or a another "low-level" exception. We do not really know what
-            # was already read from the underlying socket, so it is not even safe to retry
-            # here using the same stream. The only safe thing to do is to abort.
-            # (See also GitHub #4133).
+        except BaseException:
+            # Some OSError, CancelledError or a another "low-level" exception.
+            # We do not really know what was already read from the underlying
+            # socket, so it is not even safe to retry here using the same stream.
+            # The only safe thing to do is to abort.
+            # (See also GitHub #4133, #6548).
             self.abort()
             raise
         else:
@@ -268,7 +290,7 @@ class TCP(Comm):
         frames_nbytes = [nbytes(header), *frames_nbytes]
         frames_nbytes_total += frames_nbytes[0]
 
-        if frames_nbytes_total < 2 ** 17:  # 128kiB
+        if frames_nbytes_total < 2**17:  # 128kiB
             # small enough, send in one go
             frames = [b"".join(frames)]
             frames_nbytes = [frames_nbytes_total]
@@ -277,16 +299,25 @@ class TCP(Comm):
             # trick to enque all frames for writing beforehand
             for each_frame_nbytes, each_frame in zip(frames_nbytes, frames):
                 if each_frame_nbytes:
-                    if stream._write_buffer is None:
-                        raise StreamClosedError()
+                    # Make sure that `len(data) == data.nbytes`
+                    # See <https://github.com/tornadoweb/tornado/pull/2996>
+                    each_frame = ensure_memoryview(each_frame)
+                    for i, j in sliding_window(
+                        2,
+                        range(
+                            0,
+                            each_frame_nbytes + OPENSSL_MAX_CHUNKSIZE,
+                            OPENSSL_MAX_CHUNKSIZE,
+                        ),
+                    ):
+                        chunk = each_frame[i:j]
+                        chunk_nbytes = chunk.nbytes
 
-                    if isinstance(each_frame, memoryview):
-                        # Make sure that `len(data) == data.nbytes`
-                        # See <https://github.com/tornadoweb/tornado/pull/2996>
-                        each_frame = memoryview(each_frame).cast("B")
+                        if stream._write_buffer is None:
+                            raise StreamClosedError()
 
-                    stream._write_buffer.append(each_frame)
-                    stream._total_write_index += each_frame_nbytes
+                        stream._write_buffer.append(chunk)
+                        stream._total_write_index += chunk_nbytes
 
             # start writing frames
             stream.write(b"")
@@ -295,11 +326,14 @@ class TCP(Comm):
             self._closed = True
             if not sys.is_finalizing():
                 convert_stream_closed_error(self, e)
-        except Exception:
+        except BaseException:
             # Some OSError or a another "low-level" exception. We do not really know
             # what was already written to the underlying socket, so it is not even safe
             # to retry here using the same stream. The only safe thing to do is to
             # abort. (See also GitHub #4133).
+            # In case of, for instance, KeyboardInterrupts or other
+            # BaseExceptions that could be handled further upstream, we equally
+            # want to discard this comm
             self.abort()
             raise
 
@@ -324,14 +358,14 @@ class TCP(Comm):
                 self._finalizer.detach()
                 stream.close()
 
-    def abort(self):
+    def abort(self) -> None:
         stream, self.stream = self.stream, None
         self._closed = True
         if stream is not None and not stream.closed():
             self._finalizer.detach()
             stream.close()
 
-    def closed(self):
+    def closed(self) -> bool:
         return self._closed
 
     @property
@@ -344,8 +378,7 @@ class TLS(TCP):
     A TLS-specific version of TCP.
     """
 
-    # Workaround for OpenSSL 1.0.2 (can drop with OpenSSL 1.1.1)
-    max_shard_size = min(C_INT_MAX, TCP.max_shard_size)
+    max_shard_size = min(OPENSSL_MAX_CHUNKSIZE, TCP.max_shard_size)
 
     def _read_extra(self):
         TCP._read_extra(self)
@@ -368,7 +401,7 @@ def _expect_tls_context(connection_args):
         raise TypeError(
             "TLS expects a `ssl_context` argument of type "
             "ssl.SSLContext (perhaps check your TLS configuration?)"
-            "  Instead got %s" % str(ctx)
+            f" Instead got {ctx!r}"
         )
     return ctx
 
@@ -383,22 +416,71 @@ class RequireEncryptionMixin:
             )
 
 
-class BaseTCPConnector(Connector, RequireEncryptionMixin):
-    _executor = ThreadPoolExecutor(2, thread_name_prefix="TCP-Executor")
+_NUMERIC_ONLY = socket.AI_NUMERICHOST | socket.AI_NUMERICSERV
 
-    @property
-    def client(self):
-        # The `TCPClient` is cached on the class itself to avoid creating
-        # excess `ThreadPoolExecutor`s. We delay creation until inside an async
-        # function to avoid accessing an IOLoop from a context where a backing
-        # event loop doesn't exist.
-        cls = type(self)
-        if not hasattr(type(self), "_client"):
-            resolver = netutil.ExecutorResolver(
-                close_executor=False, executor=cls._executor
+
+async def _getaddrinfo(host, port, *, family, type=socket.SOCK_STREAM):
+    # On Python3.8 we are observing problems, particularly on slow systems
+    # For additional info, see
+    # https://github.com/dask/distributed/pull/6847#issuecomment-1208179864
+    # https://github.com/dask/distributed/pull/6847
+    # https://github.com/dask/distributed/issues/6896
+    # https://github.com/dask/distributed/issues/6846
+    if sys.version_info >= (3, 9):
+        # If host and port are numeric, then getaddrinfo doesn't block and we
+        # can skip get_running_loop().getaddrinfo which is implemented by
+        # running in a ThreadPoolExecutor. So we try first with the
+        # _NUMERIC_ONLY flags set, and then only use the threadpool if that
+        # fails with EAI_NONAME:
+        try:
+            return socket.getaddrinfo(
+                host,
+                port,
+                family=family,
+                type=type,
+                flags=_NUMERIC_ONLY,
             )
-            cls._client = TCPClient(resolver=resolver)
-        return cls._client
+        except socket.gaierror as e:
+            if e.errno != socket.EAI_NONAME:
+                raise
+
+    # That failed; it's a real hostname. We better use a thread.
+    return await asyncio.get_running_loop().getaddrinfo(
+        host, port, family=family, type=socket.SOCK_STREAM
+    )
+
+
+class _DefaultLoopResolver(netutil.Resolver):
+    """
+    Resolver implementation using `asyncio.loop.getaddrinfo`.
+    backport from Tornado 6.2+
+    https://github.com/tornadoweb/tornado/blob/3de78b7a15ba7134917a18b0755ea24d7f8fde94/tornado/netutil.py#L416-L432
+
+    With an additional optimization based on
+    https://github.com/python-trio/trio/blob/4edfd41bd5519a2e626e87f6c6ca9fb32b90a6f4/trio/_socket.py#L125-L192
+    (Copyright Contributors to the Trio project.)
+
+    And proposed to cpython in https://github.com/python/cpython/pull/31497/
+    """
+
+    async def resolve(
+        self, host: str, port: int, family: socket.AddressFamily = socket.AF_UNSPEC
+    ) -> list[tuple[int, Any]]:
+        # On Solaris, getaddrinfo fails if the given port is not found
+        # in /etc/services and no socket type is given, so we must pass
+        # one here.  The socket type used here doesn't seem to actually
+        # matter (we discard the one we get back in the results),
+        # so the addresses we return should still be usable with SOCK_DGRAM.
+        return [
+            (fam, address)
+            for fam, _, _, _, address in await _getaddrinfo(
+                host, port, family=family, type=socket.SOCK_STREAM
+            )
+        ]
+
+
+class BaseTCPConnector(Connector, RequireEncryptionMixin):
+    client: ClassVar[TCPClient] = TCPClient(resolver=_DefaultLoopResolver())
 
     async def connect(self, address, deserialize=True, **connection_args):
         self._check_encryption(address, connection_args)
@@ -406,10 +488,18 @@ class BaseTCPConnector(Connector, RequireEncryptionMixin):
         kwargs = self._get_connect_args(**connection_args)
 
         try:
-            stream = await self.client.connect(
-                ip, port, max_buffer_size=MAX_BUFFER_SIZE, **kwargs
-            )
-            # Under certain circumstances tornado will have a closed connnection with an
+            # server_hostname option (for SNI) only works with tornado.iostream.IOStream
+            if "server_hostname" in kwargs:
+                stream = await self.client.connect(
+                    ip, port, max_buffer_size=MAX_BUFFER_SIZE
+                )
+                stream = await stream.start_tls(False, **kwargs)
+            else:
+                stream = await self.client.connect(
+                    ip, port, max_buffer_size=MAX_BUFFER_SIZE, **kwargs
+                )
+
+            # Under certain circumstances tornado will have a closed connection with an
             # error and not raise a StreamClosedError.
             #
             # This occurs with tornado 5.x and openssl 1.1+
@@ -450,8 +540,10 @@ class TLSConnector(BaseTCPConnector):
     encrypted = True
 
     def _get_connect_args(self, **connection_args):
-        ctx = _expect_tls_context(connection_args)
-        return {"ssl_options": ctx}
+        tls_args = {"ssl_options": _expect_tls_context(connection_args)}
+        if connection_args.get("server_hostname"):
+            tls_args["server_hostname"] = connection_args["server_hostname"]
+        return tls_args
 
 
 class BaseTCPListener(Listener, RequireEncryptionMixin):
@@ -479,7 +571,7 @@ class BaseTCPListener(Listener, RequireEncryptionMixin):
         self.tcp_server = TCPServer(max_buffer_size=MAX_BUFFER_SIZE, **self.server_args)
         self.tcp_server.handle_stream = self._handle_stream
         backlog = int(dask.config.get("distributed.comm.socket-backlog"))
-        for i in range(5):
+        for _ in range(5):
             try:
                 # When shuffling data between workers, there can
                 # really be O(cluster size) connection requests
