@@ -5,23 +5,35 @@ See :ref:`communications` for more.
 
 .. _UCX: https://github.com/openucx/ucx
 """
+from __future__ import annotations
+
 import functools
 import logging
 import os
 import struct
-import warnings
 import weakref
-from typing import TYPE_CHECKING
+from collections.abc import Awaitable, Callable, Collection
+from typing import TYPE_CHECKING, Any
+from unittest.mock import patch
 
 import dask
 from dask.utils import parse_bytes
 
-from ..diagnostics.nvml import has_cuda_context
-from ..utils import ensure_ip, get_ip, get_ipv6, log_errors, nbytes
-from .addressing import parse_host_port, unparse_host_port
-from .core import Comm, CommClosedError, Connector, Listener
-from .registry import Backend, backends
-from .utils import ensure_concrete_host, from_frames, to_frames
+from distributed.comm.addressing import parse_host_port, unparse_host_port
+from distributed.comm.core import Comm, CommClosedError, Connector, Listener
+from distributed.comm.registry import Backend, backends
+from distributed.comm.utils import (
+    ensure_concrete_host,
+    from_frames,
+    host_array,
+    to_frames,
+)
+from distributed.diagnostics.nvml import (
+    CudaDeviceInfo,
+    get_device_index_and_uuid,
+    has_cuda_context,
+)
+from distributed.utils import ensure_ip, get_ip, get_ipv6, log_errors, nbytes
 
 logger = logging.getLogger(__name__)
 
@@ -32,21 +44,47 @@ logger = logging.getLogger(__name__)
 if TYPE_CHECKING:
     try:
         import ucp
-        from ucp import create_endpoint as ucx_create_endpoint
-        from ucp import create_listener as ucx_create_listener
     except ImportError:
         pass
 else:
-    ucp = None  # type: ignore
-    ucx_create_endpoint = None  # type: ignore
-    ucx_create_listener = None  # type: ignore
+    ucp = None
 
-host_array = None
 device_array = None
 pre_existing_cuda_context = False
 cuda_context_created = False
 
 
+_warning_suffix = (
+    "This is often the result of a CUDA-enabled library calling a CUDA runtime function before "
+    "Dask-CUDA can spawn worker processes. Please make sure any such function calls don't happen "
+    "at import time or in the global scope of a program."
+)
+
+
+def _get_device_and_uuid_str(device_info: CudaDeviceInfo) -> str:
+    return f"{device_info.device_index} ({str(device_info.uuid)})"
+
+
+def _warn_existing_cuda_context(device_info: CudaDeviceInfo, pid: int) -> None:
+    device_uuid_str = _get_device_and_uuid_str(device_info)
+    logger.warning(
+        f"A CUDA context for device {device_uuid_str} already exists "
+        f"on process ID {pid}. {_warning_suffix}"
+    )
+
+
+def _warn_cuda_context_wrong_device(
+    device_info_expected: CudaDeviceInfo, device_info_actual: CudaDeviceInfo, pid: int
+) -> None:
+    expected_device_uuid_str = _get_device_and_uuid_str(device_info_expected)
+    actual_device_uuid_str = _get_device_and_uuid_str(device_info_actual)
+    logger.warning(
+        f"Worker with process ID {pid} should have a CUDA context assigned to device "
+        f"{expected_device_uuid_str}, but instead the CUDA context is on device "
+        f"{actual_device_uuid_str}. {_warning_suffix}"
+    )
+
+
 def synchronize_stream(stream=0):
     import numba.cuda
 
@@ -57,7 +95,7 @@ def synchronize_stream(stream=0):
 
 
 def init_once():
-    global ucp, host_array, device_array
+    global ucp, device_array
     global ucx_create_endpoint, ucx_create_listener
     global pre_existing_cuda_context, cuda_context_created
 
@@ -65,13 +103,25 @@ def init_once():
         return
 
     # remove/process dask.ucx flags for valid ucx options
-    ucx_config = _scrub_ucx_config()
+    ucx_config, ucx_environment = _prepare_ucx_config()
 
     # We ensure the CUDA context is created before initializing UCX. This can't
     # be safely handled externally because communications in Dask start before
     # preload scripts run.
-    if dask.config.get("distributed.comm.ucx.create-cuda-context") is True or (
-        "TLS" in ucx_config and "cuda_copy" in ucx_config["TLS"]
+    # Precedence:
+    # 1. external environment
+    # 2. ucx_config (high level settings passed to ucp.init)
+    # 3. ucx_environment (low level settings equivalent to environment variables)
+    ucx_tls = os.environ.get(
+        "UCX_TLS",
+        ucx_config.get("TLS", ucx_environment.get("UCX_TLS", "")),
+    )
+    if (
+        dask.config.get("distributed.comm.ucx.create-cuda-context") is True
+        # This is not foolproof, if UCX_TLS=all we might require CUDA
+        # depending on configuration of UCX, but this is better than
+        # nothing
+        or ("cuda" in ucx_tls and "^cuda" not in ucx_tls)
     ):
         try:
             import numba.cuda
@@ -80,54 +130,51 @@ def init_once():
                 "CUDA support with UCX requires Numba for context management"
             )
 
-        cuda_visible_device = int(
+        cuda_visible_device = get_device_index_and_uuid(
             os.environ.get("CUDA_VISIBLE_DEVICES", "0").split(",")[0]
         )
         pre_existing_cuda_context = has_cuda_context()
-        if pre_existing_cuda_context is not False:
-            warnings.warn(
-                f"A CUDA context for device {pre_existing_cuda_context} already exists on process "
-                f"ID {os.getpid()}. This is often the result of a CUDA-enabled library calling a "
-                "CUDA runtime function before Dask-CUDA can spawn worker processes. Please make "
-                "sure any such function calls don't happen at import time or in the global scope "
-                "of a program."
+        if pre_existing_cuda_context.has_context:
+            _warn_existing_cuda_context(
+                pre_existing_cuda_context.device_info, os.getpid()
             )
 
         numba.cuda.current_context()
 
         cuda_context_created = has_cuda_context()
         if (
-            cuda_context_created is not False
-            and cuda_context_created != cuda_visible_device
+            cuda_context_created.has_context
+            and cuda_context_created.device_info.uuid != cuda_visible_device.uuid
         ):
-            warnings.warn(
-                f"Worker with process ID {os.getpid()} should have a CUDA context assigned to "
-                f"device {cuda_visible_device}, but instead the CUDA context is on device "
-                "{cuda_context_created}. This is often the result of a CUDA-enabled library "
-                "calling a CUDA runtime function before Dask-CUDA can spawn worker processes. "
-                "Please make sure any such function calls don't happen at import time or in "
-                "the global scope of a program."
+            _warn_cuda_context_wrong_device(
+                cuda_visible_device, cuda_context_created.device_info, os.getpid()
             )
 
     import ucp as _ucp
 
     ucp = _ucp
 
-    ucp.init(options=ucx_config, env_takes_precedence=True)
+    with patch.dict(os.environ, ucx_environment):
+        # We carefully ensure that ucx_environment only contains things
+        # that don't override ucx_config or existing slots in the
+        # environment, so the user's external environment can safely
+        # override things here.
+        ucp.init(options=ucx_config, env_takes_precedence=True)
 
-    # Find the function, `host_array()`, to use when allocating new host arrays
-    try:
-        import numpy
-
-        host_array = lambda n: numpy.empty((n,), dtype="u1")
-    except ImportError:
-        host_array = lambda n: bytearray(n)
+    pool_size_str = dask.config.get("distributed.rmm.pool-size")
 
     # Find the function, `cuda_array()`, to use when allocating new CUDA arrays
     try:
         import rmm
 
-        device_array = lambda n: rmm.DeviceBuffer(size=n)
+        def device_array(n):
+            return rmm.DeviceBuffer(size=n)
+
+        if pool_size_str is not None:
+            pool_size = parse_bytes(pool_size_str)
+            rmm.reinitialize(
+                pool_allocator=True, managed_memory=False, initial_pool_size=pool_size
+            )
     except ImportError:
         try:
             import numba.cuda
@@ -138,6 +185,7 @@ def init_once():
                 return a
 
             device_array = numba_device_array
+
         except ImportError:
 
             def device_array(n):
@@ -145,28 +193,11 @@ def init_once():
                     "In order to send/recv CUDA arrays, Numba or RMM is required"
                 )
 
-    pool_size_str = dask.config.get("distributed.rmm.pool-size")
-    if pool_size_str is not None:
-        pool_size = parse_bytes(pool_size_str)
-        rmm.reinitialize(
-            pool_allocator=True, managed_memory=False, initial_pool_size=pool_size
-        )
-
-    try:
-        from ucp.endpoint_reuse import EndpointReuse
-    except ImportError:
-        ucx_create_endpoint = ucp.create_endpoint
-        ucx_create_listener = ucp.create_listener
-    else:
-        reuse_endpoints = dask.config.get("distributed.comm.ucx.reuse-endpoints")
-        if (
-            reuse_endpoints is None and ucp.get_ucx_version() >= (1, 11, 0)
-        ) or reuse_endpoints is False:
-            ucx_create_endpoint = ucp.create_endpoint
-            ucx_create_listener = ucp.create_listener
-        else:
-            ucx_create_endpoint = EndpointReuse.create_endpoint
-            ucx_create_listener = EndpointReuse.create_listener
+        if pool_size_str is not None:
+            logger.warning(
+                "Initial RMM pool size defined, but RMM is not available. "
+                "Please consider installing RMM or removing the pool size option."
+            )
 
 
 def _close_comm(ref):
@@ -215,7 +246,9 @@ class UCX(Comm):
     4. Read all the data frames.
     """
 
-    def __init__(self, ep, local_addr: str, peer_addr: str, deserialize: bool = True):
+    def __init__(  # type: ignore[no-untyped-def]
+        self, ep, local_addr: str, peer_addr: str, deserialize: bool = True
+    ):
         super().__init__(deserialize=deserialize)
         self._ep = ep
         if local_addr:
@@ -245,122 +278,138 @@ class UCX(Comm):
     def peer_address(self) -> str:
         return self._peer_addr
 
+    @log_errors
     async def write(
         self,
         msg: dict,
-        serializers=("cuda", "dask", "pickle", "error"),
+        serializers: Collection[str] | None = None,
         on_error: str = "message",
-    ):
-        with log_errors():
-            if self.closed():
-                raise CommClosedError("Endpoint is closed -- unable to send message")
-            try:
-                if serializers is None:
-                    serializers = ("cuda", "dask", "pickle", "error")
-                # msg can also be a list of dicts when sending batched messages
-                frames = await to_frames(
-                    msg,
-                    serializers=serializers,
-                    on_error=on_error,
-                    allow_offload=self.allow_offload,
-                )
-                nframes = len(frames)
-                cuda_frames = tuple(
-                    hasattr(f, "__cuda_array_interface__") for f in frames
-                )
-                sizes = tuple(nbytes(f) for f in frames)
-                cuda_send_frames, send_frames = zip(
-                    *(
-                        (is_cuda, each_frame)
-                        for is_cuda, each_frame in zip(cuda_frames, frames)
-                        if nbytes(each_frame) > 0
-                    )
-                )
+    ) -> int:
+        if self.closed():
+            raise CommClosedError("Endpoint is closed -- unable to send message")
 
-                # Send meta data
+        if serializers is None:
+            serializers = ("cuda", "dask", "pickle", "error")
+        # msg can also be a list of dicts when sending batched messages
+        frames = await to_frames(
+            msg,
+            serializers=serializers,
+            on_error=on_error,
+            allow_offload=self.allow_offload,
+        )
+        nframes = len(frames)
+        cuda_frames = tuple(hasattr(f, "__cuda_array_interface__") for f in frames)
+        sizes = tuple(nbytes(f) for f in frames)
+        cuda_send_frames, send_frames = zip(
+            *(
+                (is_cuda, each_frame)
+                for is_cuda, each_frame in zip(cuda_frames, frames)
+                if nbytes(each_frame) > 0
+            )
+        )
 
-                # Send close flag and number of frames (_Bool, int64)
-                await self.ep.send(struct.pack("?Q", False, nframes))
-                # Send which frames are CUDA (bool) and
-                # how large each frame is (uint64)
-                await self.ep.send(
-                    struct.pack(nframes * "?" + nframes * "Q", *cuda_frames, *sizes)
-                )
+        try:
+            # Send meta data
 
-                # Send frames
+            # Send close flag and number of frames (_Bool, int64)
+            await self.ep.send(struct.pack("?Q", False, nframes))
+            # Send which frames are CUDA (bool) and
+            # how large each frame is (uint64)
+            await self.ep.send(
+                struct.pack(nframes * "?" + nframes * "Q", *cuda_frames, *sizes)
+            )
 
-                # It is necessary to first synchronize the default stream before start
-                # sending We synchronize the default stream because UCX is not
-                # stream-ordered and syncing the default stream will wait for other
-                # non-blocking CUDA streams. Note this is only sufficient if the memory
-                # being sent is not currently in use on non-blocking CUDA streams.
-                if any(cuda_send_frames):
-                    synchronize_stream(0)
+            # Send frames
 
-                for each_frame in send_frames:
-                    await self.ep.send(each_frame)
-                return sum(sizes)
-            except (ucp.exceptions.UCXBaseException):
-                self.abort()
-                raise CommClosedError("While writing, the connection was closed")
+            # It is necessary to first synchronize the default stream before start
+            # sending We synchronize the default stream because UCX is not
+            # stream-ordered and syncing the default stream will wait for other
+            # non-blocking CUDA streams. Note this is only sufficient if the memory
+            # being sent is not currently in use on non-blocking CUDA streams.
+            if any(cuda_send_frames):
+                synchronize_stream(0)
 
+            for each_frame in send_frames:
+                await self.ep.send(each_frame)
+            return sum(sizes)
+        except (ucp.exceptions.UCXBaseException):
+            self.abort()
+            raise CommClosedError("While writing, the connection was closed")
+
+    @log_errors
     async def read(self, deserializers=("cuda", "dask", "pickle", "error")):
-        with log_errors():
-            if deserializers is None:
-                deserializers = ("cuda", "dask", "pickle", "error")
+        if deserializers is None:
+            deserializers = ("cuda", "dask", "pickle", "error")
+
+        try:
+            # Recv meta data
+
+            # Recv close flag and number of frames (_Bool, int64)
+            msg = host_array(struct.calcsize("?Q"))
+            await self.ep.recv(msg)
+            (shutdown, nframes) = struct.unpack("?Q", msg)
+
+            if shutdown:  # The writer is closing the connection
+                raise CommClosedError("Connection closed by writer")
+
+            # Recv which frames are CUDA (bool) and
+            # how large each frame is (uint64)
+            header_fmt = nframes * "?" + nframes * "Q"
+            header = host_array(struct.calcsize(header_fmt))
+            await self.ep.recv(header)
+            header = struct.unpack(header_fmt, header)
+            cuda_frames, sizes = header[:nframes], header[nframes:]
+        except BaseException as e:
+            # In addition to UCX exceptions, may be CancelledError or another
+            # "low-level" exception. The only safe thing to do is to abort.
+            # (See also https://github.com/dask/distributed/pull/6574).
+            self.abort()
+            raise CommClosedError(
+                f"Connection closed by writer.\nInner exception: {e!r}"
+            )
+        else:
+            # Recv frames
+            frames = [
+                device_array(each_size) if is_cuda else host_array(each_size)
+                for is_cuda, each_size in zip(cuda_frames, sizes)
+            ]
+            cuda_recv_frames, recv_frames = zip(
+                *(
+                    (is_cuda, each_frame)
+                    for is_cuda, each_frame in zip(cuda_frames, frames)
+                    if nbytes(each_frame) > 0
+                )
+            )
+
+            # It is necessary to first populate `frames` with CUDA arrays and synchronize
+            # the default stream before starting receiving to ensure buffers have been allocated
+            if any(cuda_recv_frames):
+                synchronize_stream(0)
 
             try:
-                # Recv meta data
-
-                # Recv close flag and number of frames (_Bool, int64)
-                msg = host_array(struct.calcsize("?Q"))
-                await self.ep.recv(msg)
-                (shutdown, nframes) = struct.unpack("?Q", msg)
-
-                if shutdown:  # The writer is closing the connection
-                    raise CommClosedError("Connection closed by writer")
-
-                # Recv which frames are CUDA (bool) and
-                # how large each frame is (uint64)
-                header_fmt = nframes * "?" + nframes * "Q"
-                header = host_array(struct.calcsize(header_fmt))
-                await self.ep.recv(header)
-                header = struct.unpack(header_fmt, header)
-                cuda_frames, sizes = header[:nframes], header[nframes:]
-            except (
-                ucp.exceptions.UCXCloseError,
-                ucp.exceptions.UCXCanceled,
-            ) + (getattr(ucp.exceptions, "UCXConnectionReset", ()),):
-                self.abort()
-                raise CommClosedError("Connection closed by writer")
-            else:
-                # Recv frames
-                frames = [
-                    device_array(each_size) if is_cuda else host_array(each_size)
-                    for is_cuda, each_size in zip(cuda_frames, sizes)
-                ]
-                cuda_recv_frames, recv_frames = zip(
-                    *(
-                        (is_cuda, each_frame)
-                        for is_cuda, each_frame in zip(cuda_frames, frames)
-                        if nbytes(each_frame) > 0
-                    )
-                )
-
-                # It is necessary to first populate `frames` with CUDA arrays and synchronize
-                # the default stream before starting receiving to ensure buffers have been allocated
-                if any(cuda_recv_frames):
-                    synchronize_stream(0)
-
                 for each_frame in recv_frames:
                     await self.ep.recv(each_frame)
+            except BaseException as e:
+                # In addition to UCX exceptions, may be CancelledError or another
+                # "low-level" exception. The only safe thing to do is to abort.
+                # (See also https://github.com/dask/distributed/pull/6574).
+                self.abort()
+                raise CommClosedError(
+                    f"Connection closed by writer.\nInner exception: {e!r}"
+                )
+
+            try:
                 msg = await from_frames(
                     frames,
                     deserialize=self.deserialize,
                     deserializers=deserializers,
                     allow_offload=self.allow_offload,
                 )
-                return msg
+            except EOFError:
+                # Frames possibly garbled or truncated by communication error
+                self.abort()
+                raise CommClosedError("Aborted stream on truncated data")
+            return msg
 
     async def close(self):
         self._closed = True
@@ -407,17 +456,15 @@ class UCXConnector(Connector):
     comm_class = UCX
     encrypted = False
 
-    async def connect(self, address: str, deserialize=True, **connection_args) -> UCX:
+    async def connect(
+        self, address: str, deserialize: bool = True, **connection_args: Any
+    ) -> UCX:
         logger.debug("UCXConnector.connect: %s", address)
         ip, port = parse_host_port(address)
         init_once()
         try:
-            ep = await ucx_create_endpoint(ip, port)
-        except (ucp.exceptions.UCXCloseError, ucp.exceptions.UCXCanceled,) + (
-            getattr(ucp.exceptions, "UCXConnectionReset", ()),
-            getattr(ucp.exceptions, "UCXNotConnected", ()),
-            getattr(ucp.exceptions, "UCXUnreachable", ()),
-        ):  # type: ignore
+            ep = await ucp.create_endpoint(ip, port)
+        except ucp.exceptions.UCXBaseException:
             raise CommClosedError("Connection closed before handshake completed")
         return self.comm_class(
             ep,
@@ -435,10 +482,10 @@ class UCXListener(Listener):
     def __init__(
         self,
         address: str,
-        comm_handler: None,
-        deserialize=False,
-        allow_offload=True,
-        **connection_args,
+        comm_handler: Callable[[UCX], Awaitable[None]] | None = None,
+        deserialize: bool = False,
+        allow_offload: bool = True,
+        **connection_args: Any,
     ):
         if not address.startswith("ucx"):
             address = "ucx://" + address
@@ -476,7 +523,7 @@ class UCXListener(Listener):
                 await self.comm_handler(ucx)
 
         init_once()
-        self.ucp_server = ucx_create_listener(serve_forever, port=self._input_port)
+        self.ucp_server = ucp.create_listener(serve_forever, port=self._input_port)
 
     def stop(self):
         self.ucp_server = None
@@ -537,19 +584,24 @@ class UCXBackend(Backend):
 backends["ucx"] = UCXBackend()
 
 
-def _scrub_ucx_config():
-    """Function to scrub dask config options for valid UCX config options"""
+def _prepare_ucx_config():
+    """Translate dask config options to appropriate UCX config options
+
+    Returns
+    -------
+    tuple
+        Options suitable for passing to ``ucp.init`` and additional
+        UCX options that will be inserted directly into the environment
+        while calling ``ucp.init``.
+    """
 
     # configuration of UCX can happen in two ways:
     # 1) high level on/off flags which correspond to UCX configuration
-    # 2) explicitly defined UCX configuration flags
+    # 2) explicitly defined UCX configuration flags in distributed.comm.ucx.environment
+    # High-level settings in (1) are preferred to settings in (2)
+    # Settings in the external environment override both
 
-    # import does not initialize ucp -- this will occur outside this function
-    from ucp import get_config, get_ucx_version
-
-    ucx_110 = get_ucx_version() >= (1, 10, 0)
-
-    options = {}
+    high_level_options = {}
 
     # if any of the high level flags are set, as long as they are not Null/None,
     # we assume we should configure basic TLS settings for UCX, otherwise we
@@ -562,11 +614,11 @@ def _scrub_ucx_config():
         ]
     ):
         if dask.config.get("distributed.comm.ucx.rdmacm"):
-            tls = "tcp" if ucx_110 else "tcp,rdmacm"
+            tls = "tcp"
             tls_priority = "rdmacm"
         else:
-            tls = "tcp" if ucx_110 else "tcp,sockcm"
-            tls_priority = "tcp" if ucx_110 else "sockcm"
+            tls = "tcp"
+            tls_priority = "tcp"
 
         # CUDA COPY can optionally be used with ucx -- we rely on the user
         # to define when messages will include CUDA objects.  Note:
@@ -584,18 +636,27 @@ def _scrub_ucx_config():
         if dask.config.get("distributed.comm.ucx.nvlink"):
             tls = tls + ",cuda_ipc"
 
-        options = {"TLS": tls, "SOCKADDR_TLS_PRIORITY": tls_priority}
+        high_level_options = {"TLS": tls, "SOCKADDR_TLS_PRIORITY": tls_priority}
 
-        net_devices = dask.config.get("distributed.comm.ucx.net-devices")
-        if net_devices is not None and net_devices != "":
-            options["NET_DEVICES"] = net_devices
-
-    # ANY UCX options defined in config will overwrite high level dask.ucx flags
-    valid_ucx_vars = list(get_config().keys())
-    for k, v in options.items():
-        if k not in valid_ucx_vars:
-            logger.debug(
-                f"Key: {k} with value: {v} not a valid UCX configuration option"
+    # Pick up any other ucx environment settings
+    environment_options = {}
+    for k, v in dask.config.get("distributed.comm.ucx.environment", {}).items():
+        # {"some-name": value} is translated to {"UCX_SOME_NAME": value}
+        key = "_".join(map(str.upper, ("UCX", *k.split("-"))))
+        if (hl_key := key[4:]) in high_level_options:
+            logger.warning(
+                f"Ignoring {k}={v} ({key=}) in ucx.environment, "
+                f"preferring {hl_key}={high_level_options[hl_key]} "
+                "from high level options"
             )
+        elif key in os.environ:
+            # This is only info because setting UCX configuration via
+            # environment variables is a reasonably common approach
+            logger.info(
+                f"Ignoring {k}={v} ({key=}) in ucx.environment, "
+                f"preferring {key}={os.environ[key]} from external environment"
+            )
+        else:
+            environment_options[key] = v
 
-    return options
+    return high_level_options, environment_options
