@@ -1,7 +1,8 @@
+from __future__ import annotations
+
 import functools
 import warnings
 
-import bokeh
 from bokeh.application import Application
 from bokeh.application.handlers.function import FunctionHandler
 from bokeh.server.server import BokehTornado
@@ -10,15 +11,30 @@ from packaging.version import parse as parse_version
 
 import dask
 
-if parse_version(bokeh.__version__) < parse_version("2.1.1"):
+from distributed.dashboard.utils import BOKEH_VERSION
+from distributed.versions import MAX_BOKEH_VERSION, MIN_BOKEH_VERSION
+
+if BOKEH_VERSION < parse_version(MIN_BOKEH_VERSION) or BOKEH_VERSION > parse_version(
+    MAX_BOKEH_VERSION
+):
     warnings.warn(
-        "\nDask needs bokeh >= 2.1.1 for the dashboard."
+        f"\nDask needs bokeh >= {MIN_BOKEH_VERSION}, < 3 for the dashboard."
+        f"\nYou have bokeh=={BOKEH_VERSION}."
         "\nContinuing without the dashboard."
     )
-    raise ImportError("Dask needs bokeh >= 2.1.1")
+    raise ImportError(
+        f"Dask needs bokeh >= {MIN_BOKEH_VERSION}, < 3, not bokeh=={BOKEH_VERSION}"
+    )
 
 
-def BokehApplication(applications, server, prefix="/", template_variables={}):
+if BOKEH_VERSION.major < 3:
+    from bokeh.models import Panel as TabPanel  # noqa: F401
+else:
+    from bokeh.models import TabPanel  # noqa: F401
+
+
+def BokehApplication(applications, server, prefix="/", template_variables=None):
+    template_variables = template_variables or {}
     prefix = "/" + prefix.strip("/") + "/" if prefix else "/"
 
     extra = {"prefix": prefix, **template_variables}
