@@ -6,11 +6,12 @@ import sys
 import warnings
 import weakref
 from json import dumps
+from typing import Any
 
 import dask
 import dask.config
 
-from .spec import ProcessInterface, SpecCluster
+from distributed.deploy.spec import ProcessInterface, SpecCluster
 
 logger = logging.getLogger(__name__)
 
@@ -37,13 +38,15 @@ class Process(ProcessInterface):
         await super().start()
 
     async def close(self):
-        self.proc.kill()  # https://github.com/ronf/asyncssh/issues/112
-        self.connection.close()
+        if self.proc:
+            self.proc.kill()  # https://github.com/ronf/asyncssh/issues/112
+        if self.connection:
+            self.connection.close()
         await super().close()
 
 
 class Worker(Process):
-    """A Remote Dask Worker controled by SSH
+    """A Remote Dask Worker controlled by SSH
 
     Parameters
     ----------
@@ -58,11 +61,11 @@ class Worker(Process):
     remote_python: str
         Path to Python on remote node to run this worker.
     kwargs: dict
-        These will be passed through the dask-worker CLI to the
+        These will be passed through the dask worker CLI to the
         dask.distributed.Worker class
     """
 
-    def __init__(
+    def __init__(  # type: ignore[no-untyped-def]
         self,
         scheduler: str,
         address: str,
@@ -89,7 +92,37 @@ class Worker(Process):
         self.kwargs = copy.copy(kwargs)
         self.name = name
         self.remote_python = remote_python
-        self.nprocs = self.kwargs.pop("nprocs", 1)
+        if kwargs.get("nprocs") is not None and kwargs.get("n_workers") is not None:
+            raise ValueError(
+                "Both nprocs and n_workers were specified. Use n_workers only."
+            )
+        elif kwargs.get("nprocs") is not None:
+            warnings.warn(
+                "The nprocs argument will be removed in a future release. It has been "
+                "renamed to n_workers.",
+                FutureWarning,
+            )
+            self.n_workers = self.kwargs.pop("nprocs", 1)
+        else:
+            self.n_workers = self.kwargs.pop("n_workers", 1)
+
+    @property
+    def nprocs(self):
+        warnings.warn(
+            "The nprocs attribute will be removed in a future release. It has been "
+            "renamed to n_workers.",
+            FutureWarning,
+        )
+        return self.n_workers
+
+    @nprocs.setter
+    def nprocs(self, value):
+        warnings.warn(
+            "The nprocs attribute will be removed in a future release. It has been "
+            "renamed to n_workers.",
+            FutureWarning,
+        )
+        self.n_workers = value
 
     async def start(self):
         try:
@@ -138,7 +171,7 @@ class Worker(Process):
                                 **self.kwargs,
                             },
                         }
-                        for i in range(self.nprocs)
+                        for i in range(self.n_workers)
                     }
                 ),
             ]
@@ -148,7 +181,7 @@ class Worker(Process):
 
         # We watch stderr in order to get the address, then we return
         started_workers = 0
-        while started_workers < self.nprocs:
+        while started_workers < self.n_workers:
             line = await self.proc.stderr.readline()
             if not line.strip():
                 raise Exception("Worker failed to start")
@@ -171,19 +204,23 @@ class Scheduler(Process):
     remote_python: str
         Path to Python on remote node to run this scheduler.
     kwargs: dict
-        These will be passed through the dask-scheduler CLI to the
+        These will be passed through the dask scheduler CLI to the
         dask.distributed.Scheduler class
     """
 
     def __init__(
-        self, address: str, connect_options: dict, kwargs: dict, remote_python=None
+        self,
+        address: str,
+        connect_options: dict,
+        kwargs: dict,
+        remote_python: str | None = None,
     ):
         super().__init__()
 
         self.address = address
         self.kwargs = kwargs
         self.connect_options = connect_options
-        self.remote_python = remote_python
+        self.remote_python = remote_python or sys.executable
 
     async def start(self):
         try:
@@ -214,9 +251,6 @@ class Scheduler(Process):
                     "Scheduler failed to set DASK_INTERNAL_INHERIT_CONFIG variable "
                 )
 
-        if not self.remote_python:
-            self.remote_python = sys.executable
-
         cmd = " ".join(
             [
                 set_env,
@@ -248,6 +282,7 @@ old_cluster_kwargs = {
     "worker_addrs",
     "nthreads",
     "nprocs",
+    "n_workers",
     "ssh_username",
     "ssh_port",
     "ssh_private_key",
@@ -263,14 +298,14 @@ old_cluster_kwargs = {
 
 def SSHCluster(
     hosts: list[str] | None = None,
-    connect_options: dict | list[dict] = {},
-    worker_options: dict = {},
-    scheduler_options: dict = {},
+    connect_options: dict | list[dict] | None = None,
+    worker_options: dict | None = None,
+    scheduler_options: dict | None = None,
     worker_module: str = "deprecated",
     worker_class: str = "distributed.Nanny",
     remote_python: str | list[str] | None = None,
-    **kwargs,
-):
+    **kwargs: Any,
+) -> SpecCluster:
     """Deploy a Dask cluster using SSH
 
     The SSHCluster function deploys a Dask Scheduler and Workers for you on a
@@ -292,22 +327,22 @@ def SSHCluster(
 
     Parameters
     ----------
-    hosts : list[str]
+    hosts
         List of hostnames or addresses on which to launch our cluster.
         The first will be used for the scheduler and the rest for workers.
-    connect_options : dict or list of dict, optional
+    connect_options
         Keywords to pass through to :func:`asyncssh.connect`.
         This could include things such as ``port``, ``username``, ``password``
         or ``known_hosts``. See docs for :func:`asyncssh.connect` and
         :class:`asyncssh.SSHClientConnectionOptions` for full information.
         If a list it must have the same length as ``hosts``.
-    worker_options : dict, optional
+    worker_options
         Keywords to pass on to workers.
-    scheduler_options : dict, optional
+    scheduler_options
         Keywords to pass on to scheduler.
-    worker_class: str
+    worker_class
         The python class to use to create the worker(s).
-    remote_python : str or list of str, optional
+    remote_python
         Path to Python on remote nodes.
 
     Examples
@@ -336,7 +371,7 @@ def SSHCluster(
     >>> cluster = SSHCluster(
     ...     ["localhost", "localhost", "localhost", "localhost"],
     ...     connect_options={"known_hosts": None},
-    ...     worker_options={"nthreads": 2, "nprocs": 2},
+    ...     worker_options={"nthreads": 2, "n_workers": 2},
     ...     scheduler_options={"port": 0, "dashboard_address": ":8797"}
     ... )
     >>> client = Client(cluster)
@@ -358,6 +393,10 @@ def SSHCluster(
     dask.distributed.Worker
     asyncssh.connect
     """
+    connect_options = connect_options or {}
+    worker_options = worker_options or {}
+    scheduler_options = scheduler_options or {}
+
     if worker_module != "deprecated":
         raise ValueError(
             "worker_module has been deprecated in favor of worker_class. "
@@ -365,7 +404,7 @@ def SSHCluster(
         )
 
     if set(kwargs) & old_cluster_kwargs:
-        from .old_ssh import SSHCluster as OldSSHCluster
+        from distributed.deploy.old_ssh import SSHCluster as OldSSHCluster
 
         warnings.warn(
             "Note that the SSHCluster API has been replaced.  "
@@ -373,7 +412,7 @@ def SSHCluster(
             "This will be removed in the future"
         )
         kwargs.setdefault("worker_addrs", hosts)
-        return OldSSHCluster(**kwargs)
+        return OldSSHCluster(**kwargs)  # type: ignore
 
     if not hosts:
         raise ValueError(
