@@ -31,9 +31,7 @@ Known issues
   flag would solve this issue. See pytest_rerunfailures code for inspiration.
 
 - The @gen_cluster fixture leaks 2 fds on the first test decorated with it within a test
-  suite; this is likely caused by an incomplete warmup routine of
-  distributed.comm.tcp.BaseTCPConnector.
-  This issue would also be fixed by rerunning failing tests.
+  suite; This issue would also be fixed by rerunning failing tests.
 
 - The @pytest.mark.flaky decorator (pytest_rerunfailures) completely disables this
   plugin for the decorated tests.
@@ -57,9 +55,7 @@ from typing import Any, ClassVar
 import psutil
 import pytest
 
-from .comm.tcp import BaseTCPConnector
-from .compatibility import WINDOWS
-from .metrics import time
+from distributed.metrics import time
 
 
 def pytest_addoption(parser):
@@ -86,7 +82,7 @@ def pytest_addoption(parser):
     )
 
 
-def pytest_configure(config) -> None:
+def pytest_configure(config: pytest.Config) -> None:
     leaks = config.getvalue("leaks")
     if not leaks:
         return
@@ -157,14 +153,13 @@ class DemoChecker(ResourceChecker, name="demo"):
 
 
 class FDChecker(ResourceChecker, name="fds"):
-    def __init__(self):
-        BaseTCPConnector.warmup()
-
     def measure(self) -> int:
-        if WINDOWS:
+        # Note: WINDOWS constant doesn't work with `mypy --platform win32`
+        if sys.platform == "win32":
             # Don't use num_handles(); you'll get tens of thousands of reported leaks
             return 0
-        return psutil.Process().num_fds()
+        else:
+            return psutil.Process().num_fds()
 
     def has_leak(self, before: int, after: int) -> bool:
         return after > before
@@ -187,9 +182,6 @@ class RSSMemoryChecker(ResourceChecker, name="memory"):
 
 
 class ActiveThreadsChecker(ResourceChecker, name="threads"):
-    def __init__(self):
-        BaseTCPConnector.warmup()
-
     def measure(self) -> set[threading.Thread]:
         return set(threading.enumerate())
 
@@ -304,7 +296,7 @@ class TracemallocMemoryChecker(ResourceChecker, name="tracemalloc"):
         self,
         before: tuple[int, tracemalloc.Snapshot],
         after: tuple[int, tracemalloc.Snapshot],
-    ):
+    ) -> bool:
         return after[0] > before[0] + self.LEAK_THRESHOLD
 
     def format(
