@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import logging
 
 from dask.sizeof import sizeof
@@ -6,7 +8,7 @@ from dask.utils import format_bytes
 logger = logging.getLogger(__name__)
 
 
-def safe_sizeof(obj, default_size: float = 1e6) -> int:
+def safe_sizeof(obj: object, default_size: float = 1e6) -> int:
     """Safe variant of sizeof that captures and logs exceptions
 
     This returns a default size of 1e6 if the sizeof function fails
