@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import asyncio
 import operator
 from time import sleep
@@ -8,7 +10,7 @@ import dask
 
 from distributed import (
     Actor,
-    ActorFuture,
+    BaseActorFuture,
     Client,
     Future,
     Nanny,
@@ -16,6 +18,7 @@ from distributed import (
     get_client,
     wait,
 )
+from distributed.actor import _LateLoopEvent
 from distributed.metrics import time
 from distributed.utils_test import cluster, gen_cluster
 
@@ -39,16 +42,6 @@ class Counter:
         return self.n
 
 
-class UsesCounter:
-    # An actor whose method argument is another actor
-
-    def do_inc(self, ac):
-        return ac.increment().result()
-
-    async def ado_inc(self, ac):
-        return await ac.ainc()
-
-
 class List:
     L: list = []
 
@@ -82,13 +75,12 @@ async def test_client_actions(s, a, b, direct_to_workers):
         assert counter._address
         assert hasattr(counter, "increment")
         assert hasattr(counter, "add")
-        assert hasattr(counter, "n")
 
         assert await counter.n == 0
 
         assert counter._address == a.address
 
-        assert isinstance(a.actors[counter.key], Counter)
+        assert isinstance(a.state.actors[counter.key], Counter)
         assert s.tasks[counter.key].actor
 
         await asyncio.gather(counter.increment(), counter.increment())
@@ -113,7 +105,7 @@ async def test_worker_actions(c, s, a, b, separate_thread):
         assert counter._address == a_address
 
         future = counter.increment(separate_thread=separate_thread)
-        assert isinstance(future, ActorFuture)
+        assert isinstance(future, BaseActorFuture)
         assert "Future" in type(future).__name__
         end = future.result(timeout=1)
         assert end > start
@@ -131,7 +123,7 @@ async def test_Actor(c, s, a, b):
 
     assert counter._cls == Counter
 
-    assert hasattr(counter, "n")
+    assert await counter.n == 0
     assert hasattr(counter, "increment")
     assert hasattr(counter, "add")
 
@@ -196,7 +188,7 @@ async def test_gc(c, s, a, b):
     await wait(actor)
     del actor
 
-    while a.actors or b.actors:
+    while a.state.actors or b.state.actors:
         await asyncio.sleep(0.01)
 
 
@@ -210,7 +202,7 @@ async def test_track_dependencies(c, s, a, b):
 
     await asyncio.sleep(0.3)
 
-    assert a.actors or b.actors
+    assert a.state.actors or b.state.actors
 
 
 @gen_cluster(client=True)
@@ -218,7 +210,7 @@ async def test_future(c, s, a, b):
     counter = c.submit(Counter, actor=True, workers=[a.address])
     assert isinstance(counter, Future)
     await wait(counter)
-    assert isinstance(a.actors[counter.key], Counter)
+    assert isinstance(a.state.actors[counter.key], Counter)
 
     counter = await counter
     assert isinstance(counter, Actor)
@@ -266,6 +258,27 @@ def test_sync(client):
     assert "distributed.actor" not in repr(future)
 
 
+def test_timeout(client):
+    class Waiter:
+        def __init__(self):
+            self.event = _LateLoopEvent()
+
+        async def set(self):
+            self.event.set()
+
+        async def wait(self):
+            return await self.event.wait()
+
+    event = client.submit(Waiter, actor=True).result()
+    future = event.wait()
+
+    with pytest.raises(asyncio.TimeoutError):
+        future.result(timeout="0.001s")
+
+    event.set().result()
+    assert future.result() is True
+
+
 @gen_cluster(client=True, config={"distributed.comm.timeouts.connect": "1s"})
 async def test_failed_worker(c, s, a, b):
     future = c.submit(Counter, actor=True, workers=[a.address])
@@ -274,19 +287,15 @@ async def test_failed_worker(c, s, a, b):
 
     await a.close()
 
-    with pytest.raises(Exception) as info:
+    with pytest.raises(ValueError, match="Worker holding Actor was lost"):
         await counter.increment()
 
-    assert "actor" in str(info.value).lower()
-    assert "worker" in str(info.value).lower()
-    assert "lost" in str(info.value).lower()
-
 
 @gen_cluster(client=True)
 async def bench(c, s, a, b):
     counter = await c.submit(Counter, actor=True)
 
-    for i in range(1000):
+    for _ in range(1000):
         await counter.increment()
 
 
@@ -345,14 +354,17 @@ async def test_many_computations(c, s, a, b):
     counter = await c.submit(Counter, actor=True)
 
     def add(n, counter):
-        for i in range(n):
+        for _ in range(n):
             counter.increment().result()
 
     futures = c.map(add, range(10), counter=counter)
     done = c.submit(lambda x: None, futures)
 
     while not done.done():
-        assert len(s.processing) <= a.nthreads + b.nthreads
+        assert (
+            len([ws for ws in s.workers.values() if ws.processing])
+            <= a.state.nthreads + b.state.nthreads
+        )
         await asyncio.sleep(0.01)
 
     await done
@@ -368,7 +380,7 @@ async def test_thread_safety(c, s, a, b):
             assert self.n == 0
             self.n += 1
 
-            for i in range(20):
+            for _ in range(20):
                 sleep(0.002)
                 assert self.n == 1
             self.n = 0
@@ -417,7 +429,7 @@ async def test_load_balance_map(c, s, *workers):
     actors = c.map(Foo, range(10), y=b, actor=True)
     await wait(actors)
 
-    assert all(len(w.actors) == 2 for w in workers)
+    assert all(len(w.state.actors) == 2 for w in workers)
 
 
 @gen_cluster(client=True, nthreads=[("127.0.0.1", 1)] * 4, Worker=Nanny)
@@ -462,14 +474,12 @@ async def bench_param_server(c, s, *workers):
     print(format_time(end - start))
 
 
-@pytest.mark.slow
-@pytest.mark.flaky(reruns=10, reruns_delay=5)
-@gen_cluster(client=True, timeout=120)
+@gen_cluster(client=True)
 async def test_compute(c, s, a, b):
     @dask.delayed
     def f(n, counter):
         assert isinstance(counter, Actor)
-        for i in range(n):
+        for _ in range(n):
             counter.increment().result()
 
     @dask.delayed
@@ -491,7 +501,7 @@ def test_compute_sync(client):
     @dask.delayed
     def f(n, counter):
         assert isinstance(counter, Actor), type(counter)
-        for i in range(n):
+        for _ in range(n):
             counter.increment().result()
 
     @dask.delayed
@@ -517,7 +527,10 @@ def test_compute_sync(client):
 @gen_cluster(
     client=True,
     nthreads=[("127.0.0.1", 1)],
-    config={"distributed.worker.profile.interval": "1ms"},
+    config={
+        "distributed.worker.profile.enabled": True,
+        "distributed.worker.profile.interval": "1ms",
+    },
 )
 async def test_actors_in_profile(c, s, a):
     class Sleeper:
@@ -526,7 +539,7 @@ async def test_actors_in_profile(c, s, a):
 
     sleeper = await c.submit(Sleeper, actor=True)
 
-    for i in range(5):
+    for _ in range(5):
         await sleeper.sleep(0.200)
         if (
             list(a.profile_recent["children"])[0].startswith("sleep")
@@ -538,11 +551,9 @@ async def test_actors_in_profile(c, s, a):
 
 @gen_cluster(client=True)
 async def test_waiter(c, s, a, b):
-    from tornado.locks import Event
-
     class Waiter:
         def __init__(self):
-            self.event = Event()
+            self.event = _LateLoopEvent()
 
         async def set(self):
             self.event.set()
@@ -574,7 +585,7 @@ async def test_worker_actor_handle_is_weakref(c, s, a, b):
     del counter
 
     start = time()
-    while a.actors or b.data:
+    while a.state.actors or b.data:
         await asyncio.sleep(0.1)
         assert time() < start + 30
 
@@ -617,24 +628,65 @@ def test_worker_actor_handle_is_weakref_from_compute_sync(client):
         assert time() < start + 30
 
 
-def test_one_thread_deadlock():
-    with cluster(nworkers=2) as (cl, w):
-        client = Client(cl["address"])
+def test_one_thread_deadlock(loop):
+    class UsesCounter:
+        # An actor whose method argument is another actor
+
+        def do_inc(self, ac):
+            return ac.increment().result()
+
+    with cluster(nworkers=1) as (cl, _), Client(cl["address"], loop=loop) as client:
         ac = client.submit(Counter, actor=True).result()
-        ac2 = client.submit(UsesCounter, actor=True, workers=[ac._address]).result()
+        ac2 = client.submit(UsesCounter, actor=True).result()
 
         assert ac2.do_inc(ac).result() == 1
 
 
-@gen_cluster(client=True)
-async def test_async_deadlock(client, s, a, b):
+def test_one_thread_deadlock_timeout(loop):
+    class UsesCounter:
+        # An actor whose method argument is another actor
+
+        def do_inc(self, ac):
+            # ac.increment() returns an EagerActorFuture and so the timeout
+            # cannot expire
+            return ac.increment().result(timeout=0.001)
+
+    with cluster(nworkers=1) as (cl, _), Client(cl["address"], loop=loop) as client:
+        ac = client.submit(Counter, actor=True).result()
+        ac2 = client.submit(UsesCounter, actor=True).result()
+
+        assert ac2.do_inc(ac).result() == 1
+
+
+def test_one_thread_deadlock_sync_client(loop):
+    class UsesCounter:
+        # An actor whose method argument is another actor
+
+        def do_inc(self, ac):
+            return get_client().sync(ac.increment)
+
+    with cluster(nworkers=1) as (cl, _), Client(cl["address"], loop=loop) as client:
+        ac = client.submit(Counter, actor=True).result()
+        ac2 = client.submit(UsesCounter, actor=True).result()
+
+        assert ac2.do_inc(ac).result() == 1
+
+
+@gen_cluster(client=True, nthreads=[("127.0.0.1", 1)])
+async def test_async_deadlock(client, s, a):
+    class UsesCounter:
+        # An actor whose method argument is another actor
+
+        async def ado_inc(self, ac):
+            return await ac.ainc()
+
     ac = await client.submit(Counter, actor=True)
-    ac2 = await client.submit(UsesCounter, actor=True, workers=[ac._address])
+    ac2 = await client.submit(UsesCounter, actor=True)
 
     assert (await ac2.ado_inc(ac)) == 1
 
 
-def test_exception():
+def test_exception(loop):
     class MyException(Exception):
         pass
 
@@ -646,8 +698,7 @@ def test_exception():
         def prop(self):
             raise MyException
 
-    with cluster(nworkers=2) as (cl, w):
-        client = Client(cl["address"])
+    with cluster(nworkers=2) as (cl, w), Client(cl["address"], loop=loop) as client:
         ac = client.submit(Broken, actor=True).result()
         acfut = ac.method()
         with pytest.raises(MyException):
@@ -698,13 +749,24 @@ async def test_actor_future_awaitable(client, s, a, b):
     ac = await client.submit(Counter, actor=True)
     futures = [ac.increment() for _ in range(10)]
 
-    assert all([isinstance(future, ActorFuture) for future in futures])
+    assert all([isinstance(future, BaseActorFuture) for future in futures])
 
     out = await asyncio.gather(*futures)
     assert all([future.done() for future in futures])
     assert max(out) == 10
 
 
+@gen_cluster(client=True)
+async def test_actor_future_awaitable_deadlock(client, s, a, b):
+    ac = await client.submit(Counter, actor=True)
+    f = ac.increment()
+
+    async def coro():
+        return await f
+
+    assert await asyncio.gather(coro(), coro()) == [1, 1]
+
+
 @gen_cluster(client=True)
 async def test_serialize_with_pickle(c, s, a, b):
     class Foo:
