@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import functools
 import gc
 import os
@@ -12,16 +14,17 @@ import pytest
 
 import dask
 
+from distributed import profile
 from distributed.compatibility import WINDOWS
 from distributed.diskutils import WorkSpace
 from distributed.metrics import time
-from distributed.utils import mp_context
+from distributed.utils import get_mp_context
 from distributed.utils_test import captured_logger
 
 
 def assert_directory_contents(dir_path, expected, trials=2):
     expected = [os.path.join(dir_path, p) for p in expected]
-    for i in range(trials):
+    for _ in range(trials):
         actual = [
             os.path.join(dir_path, p)
             for p in os.listdir(dir_path)
@@ -52,7 +55,8 @@ def test_workdir_simple(tmpdir):
     a.release()
     assert_contents(["bb", "bb.dirlock"])
     del b
-    gc.collect()
+    with profile.lock:
+        gc.collect()
     assert_contents([])
 
     # Generated temporary name with a prefix
@@ -87,10 +91,12 @@ def test_two_workspaces_in_same_directory(tmpdir):
 
     del ws
     del b
-    gc.collect()
+    with profile.lock:
+        gc.collect()
     assert_contents(["aa", "aa.dirlock"], trials=5)
     del a
-    gc.collect()
+    with profile.lock:
+        gc.collect()
     assert_contents([], trials=5)
 
 
@@ -169,7 +175,7 @@ def test_locking_disabled(tmpdir):
     base_dir = str(tmpdir)
 
     with dask.config.set({"distributed.worker.use-file-locking": False}):
-        with mock.patch("distributed.diskutils.locket.lock_file") as lock_file:
+        with mock.patch("locket.lock_file") as lock_file:
             assert_contents = functools.partial(assert_directory_contents, base_dir)
 
             ws = WorkSpace(base_dir)
@@ -184,7 +190,8 @@ def test_locking_disabled(tmpdir):
             a.release()
             assert_contents(["bb"])
             del b
-            gc.collect()
+            with profile.lock:
+                gc.collect()
             assert_contents([])
 
         lock_file.assert_not_called()
@@ -221,9 +228,9 @@ def test_workspace_concurrency(tmpdir):
     """
     base_dir = str(tmpdir)
 
-    err_q = mp_context.Queue()
-    purged_q = mp_context.Queue()
-    stop_evt = mp_context.Event()
+    err_q = get_mp_context().Queue()
+    purged_q = get_mp_context().Queue()
+    stop_evt = get_mp_context().Event()
     ws = WorkSpace(base_dir)
     # Make sure purging only happens in the child processes
     ws._purge_leftovers = lambda: None
@@ -233,9 +240,9 @@ def test_workspace_concurrency(tmpdir):
     max_procs = 2 if WINDOWS else 16
 
     # Run a bunch of child processes that will try to purge concurrently
-    barrier = mp_context.Barrier(parties=max_procs + 1)
+    barrier = get_mp_context().Barrier(parties=max_procs + 1)
     processes = [
-        mp_context.Process(
+        get_mp_context().Process(
             target=_workspace_concurrency,
             args=(base_dir, purged_q, err_q, stop_evt, barrier),
         )
@@ -279,3 +286,21 @@ def test_workspace_concurrency(tmpdir):
 
     # We attempted to purge most directories at some point
     assert n_purged >= 0.5 * n_created > 0
+
+
+@pytest.mark.skipif(WINDOWS, reason="Need POSIX filesystem permissions and UIDs")
+def test_unwritable_base_dir(tmpdir):
+    os.mkdir(f"{tmpdir}/bad", mode=0o500)
+    with pytest.raises(PermissionError):
+        open(f"{tmpdir}/bad/tryme", "w")
+
+    ws = WorkSpace(f"{tmpdir}/bad")
+    assert ws.base_dir == f"{tmpdir}/bad-{os.getuid()}"
+
+    os.chmod(f"{tmpdir}/bad-{os.getuid()}", 0o500)
+    with pytest.raises(PermissionError):
+        open(f"{tmpdir}/bad-{os.getuid()}/tryme", "w")
+
+    ws = WorkSpace(f"{tmpdir}/bad")
+    assert ws.base_dir.startswith(f"{tmpdir}/bad-")
+    assert ws.base_dir != f"{tmpdir}/bad-{os.getuid()}"
