@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import asyncio
 import gc
 import logging
@@ -5,13 +7,15 @@ import multiprocessing as mp
 import os
 import random
 import sys
+import warnings
 from contextlib import suppress
-from time import sleep
 from unittest import mock
 
 import psutil
 import pytest
 
+from distributed.diagnostics.plugin import WorkerPlugin
+
 pytestmark = pytest.mark.gpu
 
 from tlz import first, valmap
@@ -20,58 +24,23 @@ from tornado.ioloop import IOLoop
 import dask
 from dask.utils import tmpfile
 
-from distributed import Nanny, Scheduler, Worker, rpc, wait, worker
+from distributed import Nanny, Scheduler, Worker, profile, rpc, wait, worker
 from distributed.compatibility import LINUX, WINDOWS
 from distributed.core import CommClosedError, Status
 from distributed.diagnostics import SchedulerPlugin
 from distributed.metrics import time
 from distributed.protocol.pickle import dumps
-from distributed.utils import TimeoutError, parse_ports
-from distributed.utils_test import captured_logger, gen_cluster, gen_test, inc
+from distributed.utils import TimeoutError, get_mp_context, parse_ports
+from distributed.utils_test import (
+    captured_logger,
+    gen_cluster,
+    gen_test,
+    raises_with_cause,
+)
 
 pytestmark = pytest.mark.ci1
 
 
-@pytest.mark.slow
-@gen_cluster(nthreads=[], timeout=120)
-async def test_nanny(s):
-    async with Nanny(s.address, nthreads=2, loop=s.loop) as n:
-        async with rpc(n.address) as nn:
-            assert n.is_alive()
-            [ws] = s.workers.values()
-            assert ws.nthreads == 2
-            assert ws.nanny == n.address
-
-            await nn.kill()
-            assert not n.is_alive()
-            start = time()
-            while n.worker_address in s.workers:
-                assert time() < start + 1
-                await asyncio.sleep(0.01)
-
-            await nn.kill()
-            assert not n.is_alive()
-            assert n.worker_address not in s.workers
-
-            await nn.instantiate()
-            assert n.is_alive()
-            [ws] = s.workers.values()
-            assert ws.nthreads == 2
-            assert ws.nanny == n.address
-
-            await nn.terminate()
-            assert not n.is_alive()
-
-
-@gen_cluster(nthreads=[])
-async def test_many_kills(s):
-    n = await Nanny(s.address, nthreads=2, loop=s.loop)
-    assert n.is_alive()
-    await asyncio.gather(*(n.kill() for _ in range(5)))
-    await asyncio.gather(*(n.kill() for _ in range(5)))
-    await n.close()
-
-
 @gen_cluster(Worker=Nanny)
 async def test_str(s, a, b):
     assert a.worker_address in str(a)
@@ -82,31 +51,31 @@ async def test_str(s, a, b):
 
 @gen_cluster(nthreads=[], client=True)
 async def test_nanny_process_failure(c, s):
-    n = await Nanny(s.address, nthreads=2)
-    first_dir = n.worker_dir
+    async with Nanny(s.address, nthreads=2) as n:
+        first_dir = n.worker_dir
 
-    assert os.path.exists(first_dir)
+        assert os.path.exists(first_dir)
 
-    ww = rpc(n.worker_address)
-    await ww.update_data(data=valmap(dumps, {"x": 1, "y": 2}))
-    pid = n.pid
-    assert pid is not None
-    with suppress(CommClosedError):
-        await c.run(os._exit, 0, workers=[n.worker_address])
+        ww = rpc(n.worker_address)
+        await ww.update_data(data=valmap(dumps, {"x": 1, "y": 2}))
+        pid = n.pid
+        assert pid is not None
+        with suppress(CommClosedError):
+            await c.run(os._exit, 0, workers=[n.worker_address])
 
-    while n.pid == pid:  # wait while process dies and comes back
-        await asyncio.sleep(0.01)
+        while n.pid == pid:  # wait while process dies and comes back
+            await asyncio.sleep(0.01)
 
-    await asyncio.sleep(1)
-    while not n.is_alive():  # wait while process comes back
-        await asyncio.sleep(0.01)
+        await asyncio.sleep(1)
+        while not n.is_alive():  # wait while process comes back
+            await asyncio.sleep(0.01)
 
-    # assert n.worker_address != original_address  # most likely
+        # assert n.worker_address != original_address  # most likely
 
-    while n.worker_address not in s.nthreads or n.worker_dir is None:
-        await asyncio.sleep(0.01)
+        while n.worker_address not in s.workers or n.worker_dir is None:
+            await asyncio.sleep(0.01)
 
-    second_dir = n.worker_dir
+        second_dir = n.worker_dir
 
     await n.close()
     assert not os.path.exists(second_dir)
@@ -118,14 +87,11 @@ async def test_nanny_process_failure(c, s):
 
 @gen_cluster(nthreads=[])
 async def test_run(s):
-    n = await Nanny(s.address, nthreads=2, loop=s.loop)
-
-    with rpc(n.address) as nn:
-        response = await nn.run(function=dumps(lambda: 1))
-        assert response["status"] == "OK"
-        assert response["result"] == 1
-
-    await n.close()
+    async with Nanny(s.address, nthreads=2) as n:
+        async with rpc(n.address) as nn:
+            response = await nn.run(function=dumps(lambda: 1))
+            assert response["status"] == "OK"
+            assert response["result"] == 1
 
 
 @pytest.mark.slow
@@ -147,12 +113,14 @@ async def test_no_hang_when_scheduler_closes(s, a, b):
     Worker=Nanny, nthreads=[("127.0.0.1", 1)], worker_kwargs={"reconnect": False}
 )
 async def test_close_on_disconnect(s, w):
-    await s.close()
+    with captured_logger("distributed.nanny") as logger:
+        await s.close()
 
-    start = time()
-    while w.status != Status.closed:
-        await asyncio.sleep(0.05)
-        assert time() < start + 9
+        start = time()
+        while w.status != Status.closed:
+            await asyncio.sleep(0.05)
+            assert time() < start + 9
+    assert "Reason: scheduler-close" in logger.getvalue()
 
 
 class Something(Worker):
@@ -182,7 +150,7 @@ async def test_nanny_death_timeout(s):
     with pytest.raises(TimeoutError):
         await w
 
-    assert w.status == Status.closed
+    assert w.status == Status.failed
 
 
 @gen_cluster(client=True, Worker=Nanny)
@@ -206,17 +174,16 @@ async def test_num_fds(s):
     proc = psutil.Process()
 
     # Warm up
-    w = await Nanny(s.address)
-    await w.close()
-    del w
-    gc.collect()
+    async with Nanny(s.address):
+        pass
+    with profile.lock:
+        gc.collect()
 
     before = proc.num_fds()
 
-    for i in range(3):
-        w = await Nanny(s.address)
-        await asyncio.sleep(0.1)
-        await w.close()
+    for _ in range(3):
+        async with Nanny(s.address):
+            await asyncio.sleep(0.1)
 
     while proc.num_fds() > before:
         print("fds:", before, proc.num_fds())
@@ -227,23 +194,21 @@ async def test_num_fds(s):
 @gen_cluster(client=True, nthreads=[])
 async def test_worker_uses_same_host_as_nanny(c, s):
     for host in ["tcp://0.0.0.0", "tcp://127.0.0.2"]:
-        n = await Nanny(s.address, host=host)
+        async with Nanny(s.address, host=host):
 
-        def func(dask_worker):
-            return dask_worker.listener.listen_address
+            def func(dask_worker):
+                return dask_worker.listener.listen_address
 
-        result = await c.run(func)
-        assert host in first(result.values())
-        await n.close()
+            result = await c.run(func)
+            assert host in first(result.values())
 
 
 @gen_test()
 async def test_scheduler_file():
     with tmpfile() as fn:
         s = await Scheduler(scheduler_file=fn, dashboard_address=":0")
-        w = await Nanny(scheduler_file=fn)
-        assert set(s.workers) == {w.worker_address}
-        await w.close()
+        async with Nanny(scheduler_file=fn) as n:
+            assert set(s.workers) == {n.worker_address}
         s.stop()
 
 
@@ -265,55 +230,25 @@ async def test_nanny_timeout(c, s, a):
 
 
 @gen_cluster(
-    nthreads=[("127.0.0.1", 1)],
+    nthreads=[("", 1)] * 8,
     client=True,
-    Worker=Nanny,
-    worker_kwargs={"memory_limit": "400 MiB"},
-)
-async def test_nanny_terminate(c, s, a):
-    def leak():
-        L = []
-        while True:
-            L.append(b"0" * 5_000_000)
-            sleep(0.01)
-
-    before = a.process.pid
-    with captured_logger(logging.getLogger("distributed.nanny")) as logger:
-        future = c.submit(leak)
-        while a.process.pid == before:
-            await asyncio.sleep(0.01)
-
-    out = logger.getvalue()
-    assert "restart" in out.lower()
-    assert "memory" in out.lower()
-
-
-@gen_cluster(
-    nthreads=[("127.0.0.1", 1)] * 8,
-    client=True,
-    Worker=Worker,
     clean_kwargs={"threads": False},
+    config={"distributed.worker.memory.pause": False},
 )
-async def test_throttle_outgoing_connections(c, s, a, *workers):
-    # But a bunch of small data on worker a
-    await c.run(lambda: logging.getLogger("distributed.worker").setLevel(logging.DEBUG))
+async def test_throttle_outgoing_transfers(c, s, a, *other_workers):
+    # Put a bunch of small data on worker a
+    logging.getLogger("distributed.worker").setLevel(logging.DEBUG)
     remote_data = c.map(
         lambda x: b"0" * 10000, range(10), pure=False, workers=[a.address]
     )
     await wait(remote_data)
 
-    def pause(dask_worker):
-        # Patch paused and memory_monitor on the one worker
-        # This is is very fragile, since a refactor of memory_monitor to
-        # remove _memory_monitoring will break this test.
-        dask_worker._memory_monitoring = True
-        dask_worker.status = Status.paused
-        dask_worker.outgoing_current_count = 2
+    a.status = Status.paused
+    a.transfer_outgoing_count = 2
 
-    await c.run(pause, workers=[a.address])
     requests = [
         await a.get_data(await w.rpc.connect(w.address), keys=[f.key], who=w.address)
-        for w in workers
+        for w in other_workers
         for f in remote_data
     ]
     await wait(requests)
@@ -322,36 +257,13 @@ async def test_throttle_outgoing_connections(c, s, a, *workers):
     assert "throttling" in wlogs.lower()
 
 
-@gen_cluster(nthreads=[], client=True)
-async def test_avoid_memory_monitor_if_zero_limit(c, s):
-    nanny = await Nanny(s.address, loop=s.loop, memory_limit=0)
-    typ = await c.run(lambda dask_worker: type(dask_worker.data))
-    assert typ == {nanny.worker_address: dict}
-    pcs = await c.run(lambda dask_worker: list(dask_worker.periodic_callbacks))
-    assert "memory" not in pcs
-    assert "memory" not in nanny.periodic_callbacks
-
-    future = c.submit(inc, 1)
-    assert await future == 2
-    await asyncio.sleep(0.02)
-
-    await c.submit(inc, 2)  # worker doesn't pause
-
-    await nanny.close()
-
-
-@gen_cluster(nthreads=[], client=True)
-async def test_scheduler_address_config(c, s):
+@gen_cluster(nthreads=[])
+async def test_scheduler_address_config(s):
     with dask.config.set({"scheduler-address": s.address}):
-        nanny = await Nanny(loop=s.loop)
-        assert nanny.scheduler.address == s.address
-
-        start = time()
-        while not s.workers:
-            await asyncio.sleep(0.1)
-            assert time() < start + 10
-
-    await nanny.close()
+        async with Nanny() as nanny:
+            assert nanny.scheduler.address == s.address
+            while not s.workers:
+                await asyncio.sleep(0.01)
 
 
 @pytest.mark.slow
@@ -370,8 +282,8 @@ async def test_wait_for_scheduler():
 
 @gen_cluster(nthreads=[], client=True)
 async def test_environment_variable(c, s):
-    a = Nanny(s.address, loop=s.loop, memory_limit=0, env={"FOO": "123"})
-    b = Nanny(s.address, loop=s.loop, memory_limit=0, env={"FOO": "456"})
+    a = Nanny(s.address, memory_limit=0, env={"FOO": "123"})
+    b = Nanny(s.address, memory_limit=0, env={"FOO": "456"})
     await asyncio.gather(a, b)
     results = await c.run(lambda: os.environ["FOO"])
     assert results == {a.worker_address: "123", b.worker_address: "456"}
@@ -383,7 +295,7 @@ async def test_environment_variable_by_config(c, s, monkeypatch):
 
     with dask.config.set({"distributed.nanny.environ": "456"}):
         with pytest.raises(TypeError, match="configuration must be of type dict"):
-            Nanny(s.address, loop=s.loop, memory_limit=0)
+            Nanny(s.address, memory_limit=0)
 
     with dask.config.set({"distributed.nanny.environ": {"FOO": "456"}}):
 
@@ -391,10 +303,10 @@ async def test_environment_variable_by_config(c, s, monkeypatch):
         # kwargs > env var > config
 
         with mock.patch.dict(os.environ, {"FOO": "BAR"}, clear=True):
-            a = Nanny(s.address, loop=s.loop, memory_limit=0, env={"FOO": "123"})
-            x = Nanny(s.address, loop=s.loop, memory_limit=0)
+            a = Nanny(s.address, memory_limit=0, env={"FOO": "123"})
+            x = Nanny(s.address, memory_limit=0)
 
-        b = Nanny(s.address, loop=s.loop, memory_limit=0)
+        b = Nanny(s.address, memory_limit=0)
 
         await asyncio.gather(a, b, x)
         results = await c.run(lambda: os.environ["FOO"])
@@ -421,23 +333,57 @@ async def test_environment_variable_config(c, s, monkeypatch):
         assert results[n.worker_address]["D"] == "123"
 
 
-@gen_cluster(nthreads=[], client=True)
-async def test_data_types(c, s):
-    w = await Nanny(s.address, data=dict)
-    r = await c.run(lambda dask_worker: type(dask_worker.data))
-    assert r[w.worker_address] == dict
-    await w.close()
+@gen_cluster(
+    nthreads=[("", 1)],
+    client=True,
+    Worker=Nanny,
+    config={
+        "distributed.nanny.pre-spawn-environ": {"PRE-SPAWN": 1},
+        "distributed.nanny.environ": {"POST-SPAWN": 2},
+    },
+)
+async def test_environment_variable_pre_post_spawn(c, s, n):
+    assert n.env == {"PRE-SPAWN": "1", "POST-SPAWN": "2"}
+    results = await c.run(lambda: os.environ)
+    assert results[n.worker_address]["PRE-SPAWN"] == "1"
+    assert results[n.worker_address]["POST-SPAWN"] == "2"
+
+    del os.environ["PRE-SPAWN"]
+    assert "POST-SPAWN" not in os.environ
+
+
+@gen_cluster(client=True, nthreads=[])
+async def test_config_param_overlays(c, s):
+    with dask.config.set({"test123.foo": 1, "test123.bar": 2}):
+        async with Nanny(s.address, config={"test123.bar": 3, "test123.baz": 4}) as n:
+            out = await c.submit(lambda: dask.config.get("test123"))
+
+    assert out == {"foo": 1, "bar": 3, "baz": 4}
 
 
 @gen_cluster(nthreads=[])
 async def test_local_directory(s):
     with tmpfile() as fn:
         with dask.config.set(temporary_directory=fn):
-            w = await Nanny(s.address)
-            assert w.local_directory.startswith(fn)
-            assert "dask-worker-space" in w.local_directory
-            assert w.process.worker_dir.count("dask-worker-space") == 1
-            await w.close()
+            async with Nanny(s.address) as n:
+                assert n.local_directory.startswith(fn)
+                assert "dask-worker-space" in n.local_directory
+                assert n.process.worker_dir.count("dask-worker-space") == 1
+
+
+@pytest.mark.skipif(WINDOWS, reason="Need POSIX filesystem permissions and UIDs")
+@gen_cluster(nthreads=[])
+async def test_unwriteable_dask_worker_space(s, tmpdir):
+    os.mkdir(f"{tmpdir}/dask-worker-space", mode=0o500)
+    with pytest.raises(PermissionError):
+        open(f"{tmpdir}/dask-worker-space/tryme", "w")
+
+    with dask.config.set(temporary_directory=tmpdir):
+        async with Nanny(s.address) as n:
+            assert n.local_directory == os.path.join(
+                tmpdir, f"dask-worker-space-{os.getuid()}"
+            )
+            assert n.process.worker_dir.count(f"dask-worker-space-{os.getuid()}") == 1
 
 
 def _noop(x):
@@ -507,9 +453,9 @@ async def test_lifetime(s):
 
 
 @gen_cluster(client=True, nthreads=[])
-async def test_nanny_closes_cleanly_2(c, s):
+async def test_nanny_closes_cleanly_if_worker_is_terminated(c, s):
     async with Nanny(s.address) as n:
-        with c.rpc(n.worker_address) as w:
+        async with c.rpc(n.worker_address) as w:
             IOLoop.current().add_callback(w.terminate)
             start = time()
             while n.status != Status.closed:
@@ -534,8 +480,11 @@ async def test_nanny_port_range(c, s):
         assert n1.port == 9867  # Selects first port in range
         async with Nanny(s.address, port=nanny_port, worker_port=worker_port) as n2:
             assert n2.port == 9868  # Selects next port in range
-            with pytest.raises(
-                ValueError, match="Could not start Nanny"
+            with raises_with_cause(
+                RuntimeError,
+                "Nanny failed to start.",
+                ValueError,
+                "with port 9867:9868",
             ):  # No more ports left
                 async with Nanny(s.address, port=nanny_port, worker_port=worker_port):
                     pass
@@ -559,8 +508,8 @@ class KeyboardInterruptWorker(worker.Worker):
 
 
 @pytest.mark.parametrize("protocol", ["tcp", "ucx"])
-@pytest.mark.asyncio
-async def test_nanny_closed_by_keyboard_interrupt(cleanup, protocol):
+@gen_test()
+async def test_nanny_closed_by_keyboard_interrupt(ucx_loop, protocol):
     if protocol == "ucx":  # Skip if UCX isn't available
         pytest.importorskip("ucp")
 
@@ -568,27 +517,34 @@ async def test_nanny_closed_by_keyboard_interrupt(cleanup, protocol):
         async with Nanny(
             s.address, nthreads=1, worker_class=KeyboardInterruptWorker
         ) as n:
-            n.auto_restart = False
             await n.process.stopped.wait()
             # Check that the scheduler has been notified about the closed worker
-            assert len(s.workers) == 0
-
-
-class StartException(Exception):
-    pass
+            assert "remove-worker" in str(s.events)
 
 
 class BrokenWorker(worker.Worker):
-    async def start(self):
-        raise StartException("broken")
+    async def start_unsafe(self):
+        raise ValueError("broken")
 
 
 @gen_cluster(nthreads=[])
 async def test_worker_start_exception(s):
-    # make sure this raises the right Exception:
-    with pytest.raises(StartException):
-        async with Nanny(s.address, worker_class=BrokenWorker) as n:
-            pass
+    nanny = Nanny(s.address, worker_class=BrokenWorker)
+    with captured_logger(logger="distributed.nanny", level=logging.WARNING) as logs:
+        with raises_with_cause(
+            RuntimeError,
+            "Nanny failed to start",
+            RuntimeError,
+            "BrokenWorker failed to start",
+        ):
+            async with nanny:
+                pass
+    assert nanny.status == Status.failed
+    # ^ NOTE: `Nanny.close` sets it to `closed`, then `Server.start._close_on_failure` sets it to `failed`
+    assert nanny.process is None
+    assert "Restarting worker" not in logs.getvalue()
+    # Avoid excessive spewing. (It's also printed once extra within the subprocess, which is okay.)
+    assert logs.getvalue().count("ValueError: broken") == 1, logs.getvalue()
 
 
 @gen_cluster(nthreads=[])
@@ -616,10 +572,8 @@ async def test_environ_plugin(c, s, a, b):
 @pytest.mark.parametrize(
     "modname",
     [
-        pytest.param(
-            "numpy",
-            marks=pytest.mark.xfail(reason="distributed#5723, distributed#5729"),
-        ),
+        # numpy is always imported, and for a good reason:
+        # https://github.com/dask/distributed/issues/5729
         "scipy",
         pytest.param("pandas", marks=pytest.mark.xfail(reason="distributed#5723")),
     ],
@@ -639,3 +593,170 @@ async def test_no_unnecessary_imports_on_worker(c, s, a, modname):
 
     await c.wait_for_workers(1)
     await c.run(assert_no_import)
+
+
+@pytest.mark.slow
+@gen_cluster(client=True, Worker=Nanny)
+async def test_repeated_restarts(c, s, a, b):
+    for _ in range(3):
+        await c.restart()
+        assert len(s.workers) == 2
+
+
+@pytest.mark.slow
+@gen_cluster(
+    client=True,
+    Worker=Nanny,
+    worker_kwargs={"memory_limit": "1 GiB"},
+    nthreads=[("127.0.0.1", 1)],
+)
+async def test_restart_memory(c, s, n):
+    # First kill nanny with restart
+    await c.restart()
+
+    # then kill nanny with memory
+    from dask.distributed import KilledWorker
+
+    np = pytest.importorskip("numpy")
+    s.allowed_failures = 1
+    future = c.submit(np.ones, 300_000_000, dtype="f8")
+    with pytest.raises(KilledWorker):
+        await future
+
+    while not s.workers:
+        await asyncio.sleep(0.1)
+
+
+class BlockClose(WorkerPlugin):
+    def __init__(self, close_happened):
+        self.close_happened = close_happened
+
+    async def teardown(self, worker):
+        # Never let the worker cleanly shut down, so it has to be killed
+        self.close_happened.set()
+        while True:
+            await asyncio.sleep(10)
+
+
+@pytest.mark.slow
+@gen_cluster(nthreads=[])
+async def test_close_joins(s):
+    close_happened = get_mp_context().Event()
+
+    nanny = Nanny(s.address, plugins=[BlockClose(close_happened)])
+    async with nanny:
+        p = nanny.process
+        assert p
+        close_t = asyncio.create_task(nanny.close())
+
+        while not close_happened.wait(0):
+            await asyncio.sleep(0.01)
+
+        assert not close_t.done()
+        assert nanny.status == Status.closing
+        assert nanny.process and nanny.process.status == Status.stopping
+
+        await close_t
+
+        assert nanny.status == Status.closed
+        assert not nanny.process
+
+        assert p.status == Status.stopped
+        assert not p.process
+
+
+@gen_cluster(Worker=Nanny, nthreads=[("", 1)])
+async def test_scheduler_crash_doesnt_restart(s, a):
+    # Simulate a scheduler crash by disconnecting it first
+    # (`s.close()` would tell workers to cleanly shut down)
+    bcomm = next(iter(s.stream_comms.values()))
+    bcomm.abort()
+    await s.close()
+
+    while a.status not in {Status.closing_gracefully, Status.closed}:
+        await asyncio.sleep(0.01)
+
+    await a.finished()
+    assert a.status == Status.closed
+    assert a.process is None
+
+
+@pytest.mark.slow
+@pytest.mark.skipif(not LINUX, reason="Requires GNU libc")
+@gen_cluster(
+    client=True,
+    Worker=Nanny,
+    nthreads=[("", 2)],
+    worker_kwargs={"memory_limit": "2GiB"},
+)
+async def test_malloc_trim_threshold(c, s, a):
+    """Test that the nanny sets the MALLOC_TRIM_THRESHOLD_ environment variable before
+    starting the worker process.
+
+    This test relies on these settings to work:
+
+        distributed.nanny.pre-spawn-environ.MALLOC_TRIM_THRESHOLD_: 65536
+        distributed.worker.multiprocessing-method: spawn
+
+    We're deliberately not setting them explicitly in @gen_cluster above, as we want
+    this test to trip if somebody changes distributed.yaml.
+
+    Note
+    ----
+    This test may start failing in a future Python version if CPython switches to
+    using mimalloc by default. If it does, a thorough benchmarking exercise is needed.
+    """
+    da = pytest.importorskip("dask.array")
+
+    a = da.random.random(
+        2**29 // 8,  # 0.5 GiB,
+        chunks=160 * 2**10 // 8,  # 160 kiB
+    ).persist()
+    await wait(a)
+    # Wait for heartbeat
+    while s.memory.process < 2**29:
+        await asyncio.sleep(0.01)
+    del a
+
+    # This is the delicate bit, as it relies on
+    # 1. PyMem_Free() to be quick to invoke glibc free() when memory becomes available
+    # 2. glibc free() to be quick to invoke the kernel's sbrk() when the same happens
+    #
+    # At the moment of writing, the readings are:
+    # - 122 MiB after starting a new worker
+    # - 139 MiB after computing a trivial dask.array collection
+    # - 185 MiB at the end of this test, with MALLOC_TRIM_THRESHOLD_=65536
+    # - 698 MiB at the end of this test, without MALLOC_TRIM_THRESHOLD_
+    while s.memory.process > 250 * 2**20:
+        await asyncio.sleep(0.01)
+
+
+@gen_cluster(client=True, nthreads=[])
+async def test_default_client_does_not_propagate_to_subprocess(c, s):
+    @dask.delayed
+    def run_in_thread():
+        return
+
+    def func():
+        with warnings.catch_warnings(record=True) as rec:
+            warnings.filterwarnings(
+                "once",
+                message="Running on a single-machine scheduler",
+                category=UserWarning,
+            )
+            # If no scheduler kwarg is provided, this will
+            # automatically transition to long-running
+            dask.compute(run_in_thread(), scheduler="single-threaded")
+        return rec
+
+    async with Nanny(s.address):
+        rec = await c.submit(func)
+        assert not rec
+
+
+@gen_cluster(client=True, nthreads=[], config={"test123": 456})
+async def test_worker_inherits_temp_config(c, s):
+    with dask.config.set(test123=123):
+        async with Nanny(s.address):
+            out = await c.submit(lambda: dask.config.get("test123"))
+            assert out == 123
