@@ -1,5 +1,6 @@
+from __future__ import annotations
+
 import asyncio
-from time import time
 
 import pytest
 
@@ -7,29 +8,20 @@ import dask
 from dask import delayed
 from dask.utils import stringify
 
-from distributed import Worker
+from distributed import Lock, Worker
 from distributed.client import wait
-from distributed.utils_test import gen_cluster, inc, slowadd, slowinc
-
-
-@gen_cluster(client=True, nthreads=[])
-async def test_resources(c, s):
-    assert not s.worker_resources
-    assert not s.resources
-
-    a = Worker(s.address, loop=s.loop, resources={"GPU": 2})
-    b = Worker(s.address, loop=s.loop, resources={"GPU": 1, "DB": 1})
-    await asyncio.gather(a, b)
-
-    assert s.resources == {"GPU": {a.address: 2, b.address: 1}, "DB": {b.address: 1}}
-    assert s.worker_resources == {a.address: {"GPU": 2}, b.address: {"GPU": 1, "DB": 1}}
-
-    await b.close()
-
-    assert s.resources == {"GPU": {a.address: 2}, "DB": {}}
-    assert s.worker_resources == {a.address: {"GPU": 2}}
-
-    await a.close()
+from distributed.utils_test import NO_AMM, gen_cluster, inc, lock_inc, slowadd, slowinc
+from distributed.worker_state_machine import (
+    ComputeTaskEvent,
+    DigestMetric,
+    Execute,
+    ExecuteFailureEvent,
+    ExecuteSuccessEvent,
+    FreeKeysEvent,
+    LongRunningMsg,
+    RescheduleEvent,
+    TaskFinishedMsg,
+)
 
 
 @gen_cluster(
@@ -52,12 +44,9 @@ async def test_resource_submit(c, s, a, b):
 
     assert s.get_task_status(keys=[z.key]) == {z.key: "no-worker"}
 
-    d = await Worker(s.address, loop=s.loop, resources={"C": 10})
-
-    await wait(z)
-    assert z.key in d.data
-
-    await d.close()
+    async with Worker(s.address, resources={"C": 10}) as w:
+        await wait(z)
+        assert z.key in w.data
 
 
 @gen_cluster(
@@ -87,12 +76,12 @@ async def test_submit_many_non_overlapping_2(c, s, a, b):
 
     while len(a.data) + len(b.data) < 100:
         await asyncio.sleep(0.01)
-        assert a.executing_count <= 2
-        assert b.executing_count <= 1
+        assert a.state.executing_count <= 2
+        assert b.state.executing_count <= 1
 
     await wait(futures)
-    assert a.total_resources == a.available_resources
-    assert b.total_resources == b.available_resources
+    assert a.state.total_resources == a.state.available_resources
+    assert b.state.total_resources == b.state.available_resources
 
 
 @gen_cluster(
@@ -103,7 +92,7 @@ async def test_submit_many_non_overlapping_2(c, s, a, b):
     ],
 )
 async def test_move(c, s, a, b):
-    [x] = await c._scatter([1], workers=b.address)
+    [x] = await c.scatter([1], workers=b.address)
 
     future = c.submit(inc, x, resources={"A": 1})
 
@@ -119,7 +108,7 @@ async def test_move(c, s, a, b):
     ],
 )
 async def test_dont_work_steal(c, s, a, b):
-    [x] = await c._scatter([1], workers=a.address)
+    [x] = await c.scatter([1], workers=a.address)
 
     futures = [
         c.submit(slowadd, x, i, resources={"A": 1}, delay=0.05) for i in range(10)
@@ -145,23 +134,20 @@ async def test_map(c, s, a, b):
 
 @gen_cluster(
     client=True,
-    nthreads=[
-        ("127.0.0.1", 1, {"resources": {"A": 1}}),
-        ("127.0.0.1", 1, {"resources": {"B": 1}}),
-    ],
+    nthreads=[("", 1, {"resources": {"A": 1}}), ("", 1, {"resources": {"B": 1}})],
+    config=NO_AMM,
 )
 async def test_persist(c, s, a, b):
     with dask.annotate(resources={"A": 1}):
-        x = delayed(inc)(1)
+        x = delayed(inc)(1, dask_key_name="x")
     with dask.annotate(resources={"B": 1}):
-        y = delayed(inc)(x)
+        y = delayed(inc)(x, dask_key_name="y")
 
     xx, yy = c.persist([x, y], optimize_graph=False)
-
     await wait([xx, yy])
 
-    assert x.key in a.data
-    assert y.key in b.data
+    assert set(a.data) == {"x"}
+    assert set(b.data) == {"x", "y"}
 
 
 @gen_cluster(
@@ -250,25 +236,135 @@ async def test_minimum_resource(c, s, a):
 
     while len(a.data) < 30:
         await asyncio.sleep(0.01)
-        assert a.executing_count <= 1
+        assert a.state.executing_count <= 1
 
     await wait(futures)
-    assert a.total_resources == a.available_resources
+    assert a.state.total_resources == a.state.available_resources
 
 
-@gen_cluster(client=True, nthreads=[("127.0.0.1", 2, {"resources": {"A": 1}})])
-async def test_prefer_constrained(c, s, a):
-    futures = c.map(slowinc, range(1000), delay=0.1)
-    constrained = c.map(inc, range(10), resources={"A": 1})
+@pytest.mark.parametrize("swap", [False, True])
+@pytest.mark.parametrize("p1,p2,expect_key", [(1, 0, "y"), (0, 1, "x")])
+def test_constrained_vs_ready_priority_1(ws, p1, p2, expect_key, swap):
+    """If there are both ready and constrained tasks, those with the highest priority
+    win (note: on the Worker, priorities have their sign inverted)
+    """
+    ws.available_resources = {"R": 1}
+    ws.total_resources = {"R": 1}
+    RR = {"resource_restrictions": {"R": 1}}
 
-    start = time()
-    await wait(constrained)
-    end = time()
-    assert end - start < 4
-    has_what = dict(s.has_what)
-    processing = dict(s.processing)
-    assert len(has_what) < len(constrained) + 2  # at most two slowinc's finished
-    assert s.processing[a.address]
+    ws.handle_stimulus(ComputeTaskEvent.dummy(key="clog", stimulus_id="clog"))
+
+    stimuli = [
+        ComputeTaskEvent.dummy("x", priority=(p1,), stimulus_id="s1"),
+        ComputeTaskEvent.dummy("y", priority=(p2,), **RR, stimulus_id="s2"),
+    ]
+    if swap:
+        stimuli = stimuli[::-1]  # This must be inconsequential
+
+    instructions = ws.handle_stimulus(
+        *stimuli,
+        ExecuteSuccessEvent.dummy("clog", stimulus_id="s3"),
+    )
+    assert instructions == [
+        DigestMetric(name="compute-duration", value=1.0, stimulus_id="s3"),
+        TaskFinishedMsg.match(key="clog", stimulus_id="s3"),
+        Execute(key=expect_key, stimulus_id="s3"),
+    ]
+
+
+@pytest.mark.parametrize("swap", [False, True])
+@pytest.mark.parametrize("p1,p2,expect_key", [(1, 0, "y"), (0, 1, "x")])
+def test_constrained_vs_ready_priority_2(ws, p1, p2, expect_key, swap):
+    """If there are both ready and constrained tasks, but not enough available
+    resources, priority is inconsequential - the tasks in the ready queue are picked up.
+    """
+    ws.nthreads = 2
+    ws.available_resources = {"R": 1}
+    ws.total_resources = {"R": 1}
+    RR = {"resource_restrictions": {"R": 1}}
+
+    ws.handle_stimulus(
+        ComputeTaskEvent.dummy(key="clog1", stimulus_id="clog1"),
+        ComputeTaskEvent.dummy(key="clog2", **RR, stimulus_id="clog2"),
+    )
+
+    # Test that both priorities and order are inconsequential
+    stimuli = [
+        ComputeTaskEvent.dummy("x", priority=(p1,), stimulus_id="s1"),
+        ComputeTaskEvent.dummy("y", priority=(p2,), **RR, stimulus_id="s2"),
+    ]
+    if swap:
+        stimuli = stimuli[::-1]
+
+    instructions = ws.handle_stimulus(
+        *stimuli,
+        ExecuteSuccessEvent.dummy("clog1", stimulus_id="s3"),
+    )
+    assert instructions == [
+        DigestMetric(name="compute-duration", value=1.0, stimulus_id="s3"),
+        TaskFinishedMsg.match(key="clog1", stimulus_id="s3"),
+        Execute(key="x", stimulus_id="s3"),
+    ]
+
+
+def test_constrained_tasks_respect_priority(ws):
+    ws.available_resources = {"R": 1}
+    ws.total_resources = {"R": 1}
+    RR = {"resource_restrictions": {"R": 1}}
+
+    instructions = ws.handle_stimulus(
+        ComputeTaskEvent.dummy(key="clog", **RR, stimulus_id="clog"),
+        ComputeTaskEvent.dummy(key="x1", priority=(1,), **RR, stimulus_id="s1"),
+        ComputeTaskEvent.dummy(key="x2", priority=(2,), **RR, stimulus_id="s2"),
+        ComputeTaskEvent.dummy(key="x3", priority=(0,), **RR, stimulus_id="s3"),
+        ExecuteSuccessEvent.dummy(key="clog", stimulus_id="s4"),  # start x3
+        ExecuteSuccessEvent.dummy(key="x3", stimulus_id="s5"),  # start x1
+        ExecuteSuccessEvent.dummy(key="x1", stimulus_id="s6"),  # start x2
+    )
+    assert instructions == [
+        Execute(key="clog", stimulus_id="clog"),
+        DigestMetric(name="compute-duration", value=1.0, stimulus_id="s4"),
+        TaskFinishedMsg.match(key="clog", stimulus_id="s4"),
+        Execute(key="x3", stimulus_id="s4"),
+        DigestMetric(name="compute-duration", value=1.0, stimulus_id="s5"),
+        TaskFinishedMsg.match(key="x3", stimulus_id="s5"),
+        Execute(key="x1", stimulus_id="s5"),
+        DigestMetric(name="compute-duration", value=1.0, stimulus_id="s6"),
+        TaskFinishedMsg.match(key="x1", stimulus_id="s6"),
+        Execute(key="x2", stimulus_id="s6"),
+    ]
+
+
+def test_task_cancelled_and_readded_with_resources(ws):
+    """See https://github.com/dask/distributed/issues/6710
+
+    A task is enqueued without resources, then cancelled by the client, then re-added
+    with the same key, this time with resources.
+    Test that resources are respected.
+    """
+    ws.available_resources = {"R": 1}
+    ws.total_resources = {"R": 1}
+    RR = {"resource_restrictions": {"R": 1}}
+
+    ws.handle_stimulus(
+        ComputeTaskEvent.dummy(key="clog", **RR, stimulus_id="s1"),
+        ComputeTaskEvent.dummy(key="x", stimulus_id="s2"),
+    )
+    ts = ws.tasks["x"]
+    assert ts.state == "ready"
+    assert ts in ws.ready
+    assert ts not in ws.constrained
+    assert ts.resource_restrictions == {}
+
+    ws.handle_stimulus(
+        FreeKeysEvent(keys=["x"], stimulus_id="clog"),
+        ComputeTaskEvent.dummy(key="x", **RR, stimulus_id="s2"),
+    )
+    ts = ws.tasks["x"]
+    assert ts.state == "constrained"
+    assert ts not in ws.ready
+    assert ts in ws.constrained
+    assert ts.resource_restrictions == {"R": 1}
 
 
 @pytest.mark.skip(reason="")
@@ -291,18 +387,19 @@ async def test_balance_resources(c, s, a, b):
 @gen_cluster(client=True, nthreads=[("127.0.0.1", 2)])
 async def test_set_resources(c, s, a):
     await a.set_resources(A=2)
-    assert a.total_resources["A"] == 2
-    assert a.available_resources["A"] == 2
-    assert s.worker_resources[a.address] == {"A": 2}
-
-    future = c.submit(slowinc, 1, delay=1, resources={"A": 1})
-    while a.available_resources["A"] == 2:
-        await asyncio.sleep(0.01)
+    assert a.state.total_resources["A"] == 2
+    assert a.state.available_resources["A"] == 2
+    assert s.workers[a.address].resources == {"A": 2}
+    lock = Lock()
+    async with lock:
+        future = c.submit(lock_inc, 1, lock=lock, resources={"A": 1})
+        while a.state.available_resources["A"] == 2:
+            await asyncio.sleep(0.01)
 
     await a.set_resources(A=3)
-    assert a.total_resources["A"] == 3
-    assert a.available_resources["A"] == 2
-    assert s.worker_resources[a.address] == {"A": 3}
+    assert a.state.total_resources["A"] == 3
+    assert a.state.available_resources["A"] == 2
+    assert s.workers[a.address].resources == {"A": 3}
 
 
 @gen_cluster(
@@ -345,7 +442,7 @@ async def test_dont_optimize_out(c, s, a, b):
     await c.compute(w, resources={tuple(y.__dask_keys__()): {"A": 1}})
 
     for key in map(stringify, y.__dask_keys__()):
-        assert "executing" in str(a.story(key))
+        assert "executing" in str(a.state.story(key))
 
 
 @pytest.mark.skip(reason="atop fusion seemed to break this")
@@ -364,8 +461,8 @@ async def test_full_collections(c, s, a, b):
     z = df.x + df.y  # some extra nodes in the graph
 
     await c.compute(z, resources={tuple(z.dask): {"A": 1}})
-    assert a.log
-    assert not b.log
+    assert a.state.log
+    assert not b.state.log
 
 
 @pytest.mark.parametrize(
@@ -417,3 +514,77 @@ async def test_resources_from_python_override_config(c, s, a, b):
     info = c.scheduler_info()
     for worker in [a, b]:
         assert info["workers"][worker.address]["resources"] == {"my_resources": 10}
+
+
+@pytest.mark.parametrize(
+    "done_ev_cls", [ExecuteSuccessEvent, ExecuteFailureEvent, RescheduleEvent]
+)
+def test_cancelled_with_resources(ws_with_running_task, done_ev_cls):
+    """Test transition loop of a task with resources:
+
+    executing -> cancelled -> released
+    """
+    ws = ws_with_running_task
+    assert ws.available_resources == {"R": 0}
+
+    ws.handle_stimulus(FreeKeysEvent(keys=["x"], stimulus_id="s1"))
+    assert ws.available_resources == {"R": 0}
+
+    ws.handle_stimulus(done_ev_cls.dummy("x", stimulus_id="s2"))
+    assert ws.available_resources == {"R": 1}
+    assert "x" not in ws.tasks
+
+
+@pytest.mark.parametrize(
+    "done_ev_cls", [ExecuteSuccessEvent, ExecuteFailureEvent, RescheduleEvent]
+)
+def test_resumed_with_resources(ws_with_running_task, done_ev_cls):
+    """Test transition loop of a task with resources:
+
+    executing -> cancelled -> resumed(fetch) -> (complete execution)
+    """
+    ws = ws_with_running_task
+    ws2 = "127.0.0.1:2"
+    assert ws.available_resources == {"R": 0}
+
+    ws.handle_stimulus(FreeKeysEvent(keys=["x"], stimulus_id="s1"))
+    assert ws.available_resources == {"R": 0}
+
+    ws.handle_stimulus(
+        ComputeTaskEvent.dummy("y", who_has={"x": [ws2]}, stimulus_id="s2")
+    )
+    assert ws.available_resources == {"R": 0}
+
+    ws.handle_stimulus(done_ev_cls.dummy("x", stimulus_id="s3"))
+    assert ws.available_resources == {"R": 1}
+
+
+@pytest.mark.parametrize(
+    "done_ev_cls", [ExecuteSuccessEvent, ExecuteFailureEvent, RescheduleEvent]
+)
+def test_resumed_with_different_resources(ws_with_running_task, done_ev_cls):
+    """A task with resources is cancelled and then resumed to the same state, but with
+    different resources. This is actually possible in case of manual cancellation from
+    the client, followed by resubmit.
+    """
+    ws = ws_with_running_task
+    assert ws.available_resources == {"R": 0}
+    ts = ws.tasks["x"]
+    prev_state = ts.state
+
+    ws.handle_stimulus(FreeKeysEvent(keys=["x"], stimulus_id="s1"))
+    assert ws.available_resources == {"R": 0}
+
+    instructions = ws.handle_stimulus(
+        ComputeTaskEvent.dummy("x", stimulus_id="s2", resource_restrictions={"R": 0.4})
+    )
+    if prev_state == "long-running":
+        assert instructions == [
+            LongRunningMsg(key="x", compute_duration=None, stimulus_id="s2")
+        ]
+    else:
+        assert not instructions
+    assert ws.available_resources == {"R": 0}
+
+    ws.handle_stimulus(done_ev_cls.dummy(key="x", stimulus_id="s3"))
+    assert ws.available_resources == {"R": 1}
