@@ -1,7 +1,8 @@
+from __future__ import annotations
+
 import asyncio
 import random
 import threading
-import warnings
 from collections import defaultdict
 from time import sleep
 
@@ -32,13 +33,13 @@ async def test_submit_from_worker(c, s, a, b):
             return result
 
     x, y = c.map(func, [10, 20])
-    xx, yy = await c._gather([x, y])
+    xx, yy = await c.gather([x, y])
 
     assert xx == 10 + 1 + (10 + 1) * 2
     assert yy == 20 + 1 + (20 + 1) * 2
 
     assert len(s.transition_log) > 10
-    assert len([id for id in s.wants_what if id.lower().startswith("client")]) == 1
+    assert len([id for id in s.clients if id.lower().startswith("client")]) == 1
 
 
 @gen_cluster(client=True, nthreads=[("127.0.0.1", 1)] * 2)
@@ -77,7 +78,7 @@ async def test_scatter_from_worker(c, s, a, b):
     assert result is True
 
     start = time()
-    while not all(v == 1 for v in s.nthreads.values()):
+    while not all(ws.nthreads == 1 for ws in s.workers.values()):
         await asyncio.sleep(0.1)
         assert time() < start + 5
 
@@ -205,7 +206,7 @@ def test_dont_override_default_get(loop):
         loop=loop, processes=False, set_as_default=True, dashboard_address=":0"
     ) as c:
         assert dask.base.get_scheduler() == c.get
-        for i in range(2):
+        for _ in range(2):
             b2.compute()
 
         assert dask.base.get_scheduler() == c.get
@@ -216,12 +217,13 @@ async def test_local_client_warning(c, s, a, b):
     from distributed import local_client
 
     def func(x):
-        with warnings.catch_warnings(record=True) as record:
-            with local_client() as c:
-                x = c.submit(inc, x)
-                result = x.result()
-            assert any("worker_client" in str(r.message) for r in record)
-            return result
+        with pytest.warns(
+            UserWarning, match=r"local_client has moved to worker_client"
+        ):
+            cmgr = local_client()
+
+        with cmgr as c:
+            return c.submit(inc, x).result()
 
     future = c.submit(func, 10)
     result = await future
@@ -254,13 +256,9 @@ def test_secede_without_stealing_issue_1262():
     Tests that seceding works with the Stealing extension disabled
     https://github.com/dask/distributed/issues/1262
     """
-
-    # turn off all extensions
-    extensions = []
-
     # run the loop as an inner function so all workers are closed
     # and exceptions can be examined
-    @gen_cluster(client=True, scheduler_kwargs={"extensions": extensions})
+    @gen_cluster(client=True, scheduler_kwargs={"extensions": {}})
     async def secede_test(c, s, a, b):
         def func(x):
             with worker_client() as wc:
@@ -274,8 +272,6 @@ def test_secede_without_stealing_issue_1262():
     c, s, a, b, f = secede_test()
 
     assert f == 2
-    # ensure no workers had errors
-    assert all([f.exception() is None for f in s._worker_coroutines])
 
 
 @gen_cluster(client=True)
@@ -301,6 +297,9 @@ async def test_worker_client_rejoins(c, s, a, b):
     assert result
 
 
+@pytest.mark.xfail(
+    reason="Flaky due to https://github.com/dask/distributed/issues/5915"
+)
 @gen_cluster()
 async def test_submit_different_names(s, a, b):
     # https://github.com/dask/distributed/issues/2058
